#!/usr/bin/env python3
"""


See EOF for license/metadata/notes as applicable
"""

##-- builtin imports
from __future__ import annotations

# import abc
import datetime
import enum
import functools as ftz
import itertools as itz
import logging as logmod
import pathlib as pl
import re
import time
import types
import weakref
# from copy import deepcopy
# from dataclasses import InitVar, dataclass, field
from typing import (TYPE_CHECKING, Any, Callable, ClassVar, Final, Generic,
                    Iterable, Iterator, Mapping, Match, MutableMapping,
                    Protocol, Sequence, Tuple, TypeAlias, TypeGuard, TypeVar,
                    cast, final, overload, runtime_checkable, Generator)
from uuid import UUID, uuid1

##-- end builtin imports

##-- lib imports
import more_itertools as mitz
##-- end lib imports

##-- logging
logging = logmod.getLogger(__name__)
##-- end logging

printer = logmod.getLogger("doot._printer")

import mastodon
import tomlguard
from jgdv.apis.mastodon import errors as djerror
from jgdv.key import JGDVKey

TOOT_SIZE            : Final[int]                   = doot.config.on_fail(250, int).mastodon.toot_size()
TOOT_IMAGE_SIZE      : Final[str]                   = doot.config.on_fail(8_000_000, int).mastodon.image_size()
RESOLUTION_BLACKLIST : Final[list]                  = doot.locs.image_blacklist
RESOLUTION_RE        : Final[re.Pattern]            = re.compile(r".*?([0-9]+x[0-9]+)")
TOOT_IMAGE_TYPES     : Final[list[str]]             = [".jpg", ".png", ".gif"]

##-- expansion keys
INSTANCE_KEY           : Final[DootKey]                        = DootKey.make("mastodon")
TEXT_KEY               : Final[DootKey]                        = DootKey.make("toot_text")
IMAGE_KEY              : Final[DootKey]                        = DootKey.make("toot_image")
IMAGE_DESC             : Final[DootKey]                        = DootKey.make("toot_desc")
UPDATE                 : Final[DootKey]                        = DootKey.make("update_")
FROM_KEY               : Final[DootKey]                        = DootKey.make("from")
SECRETS                : Final[DootKey]                        = DootKey.make("mastodon_secrets")
##-- end expansion keys


class MastodonSetup:
    """ Default Mastodon Setup, using secrets from doot.locs.mastodon_secrets
      loads the secrets as a tomlguard, and accesses mastodon.access_token and mastodon.url
      ensures thers an "image_temp" location
    """
    instance = None
    _toml_kwargs = [INSTANCE_KEY, FROM_KEY]

    def __call__(self, spec, task_state) -> dict|bool|None:
        data_key = INSTANCE_KEY.redirect(spec)

        if MastodonSetup.instance is None:
            printer.info("---------- Initialising Mastodon", extra={"colour": "green"})
            secrets_path = FROM_KEY.to_path(spec, task_state, chain=SECRETS)
            secrets = tomlguard.load(secrets_path)
            MastodonSetup.instance = mastodon.Mastodon(
                access_token = secrets.mastodon.access_token,
                api_base_url = secrets.mastodon.url
            )
            doot.locs.ensure("image_temp", task=task_state['_task_name'])
        else:
            printer.debug("Reusing Instance")

        return { data_key : MastodonSetup.instance }



class MastodonPost:
    """ Default Mastodon Poster  """
    _toml_kwargs = [INSTANCE_KEY, FROM_KEY, IMAGE_KEY, IMAGE_DESC, TEXT_KEY]

    def __call__(self, spec, task_state):
        instance           = INSTANCE_KEY.to_type(spec, task_state, type_=Mastodon.Mastodon)
        text               = FROM_KEY.redirect(spec, chain=TEXT_KEY).expand(spec, task_state)
        image_path         = IMAGE_KEY.to_path(spec, task_state)
        image_desc         = IMAGE_DESC.expand(spec, task_state)

        try:
            if image_path.exists():
                return self._post_image(instance, text, image_path, image_desc)
            else:
                return self._post_text(instance, text)
        except mastodon.MastodonAPIError as err:
            general, errcode, form, detail = err.args
            resolution                     = RESOLUTION_RE.match(detail) if detail else None
            if resolution and resolution in self.resolution_blacklist:
                pass
            elif errcode == 422 and form == "Unprocessable Entity" and resolution:
                with open(RESOLUTION_BLACKLIST, 'a') as f:
                    f.write("\n" + resolution[1])

            printer.error("Mastodon Resolution Failure: %s", repr(err))
            return False
        except Exception as err:
            printer.error("Mastodon Post Failed: %s", repr(err))
            return False

    def _post_text(self, instance, text):
        printer.info("Posting Text Toot: %s", text)
        if len(text) >= TOOT_SIZE:
            printer.warning("Resulting Toot too long for mastodon: %s\n%s", len(text), text)
            return False

        result = instance.status_post(text)
        return True

    def _post_image(self, instance, text, image_path, image_desc):
        printer.info("Posting Image Toot")

        assert(image_path.exists()), f"File Doesn't Exist {image_path}"
        assert(image_path.stat().st_size < TOOT_IMAGE_SIZE), "Image to large, needs to be smaller than 8MB"
        assert(image_path.suffix.lower() in TOOT_IMAGE_TYPES), "Bad Type, needs to be a jpg, png or gif"

        media_id = instance.media_post(str(image_path), description=image_desc)
        instance.status_post(text, media_ids=media_id)
        logging.debug("Image Toot Posted")
        return True

    def _handle_resolution(self, task):
        # post to mastodon
        with open(RESOLUTION_BLACKLIST, 'r') as f:
            resolution_blacklist = {x.strip() for x in f.readlines()}

        min_x, min_y = inf, inf

        if bool(resolution_blacklist):
            min_x        = min(int(res.split("x")[0]) for res in resolution_blacklist)
            min_y        = min(int(res.split("x")[1]) for res in resolution_blacklist)

        res : str    = _get_resolution(selected_file)
        res_x, res_y = res.split("x")
        res_x, res_y = int(res_x), int(res_y)
        if res in resolution_blacklist or (min_x <= res_x and min_y <= res_y):
            logging.warning("Image is too big %s: %s", selected_file, res)
            return

    def _get_resolution(self, filepath:Path) -> str:
        result = subprocess.run(["file", str(filepath)], capture_output=True, shell=False)
        if result.returncode == 0:
            res = RESOLUTION_RE.match(result.stdout.decode())
            return res[1]

        raise djerror.DootActionError("Couldn't get image resolution", filepath, result.stdout.decode(), result.stderr.decode())

    def _maybe_compress_file(self, task):
        image = task.values['image']
        logging.debug("Attempting compression of: %s", image)
        assert(isinstance(filepath, pl.Path) and filepath.exists())
        ext               = filepath.suffix
        conversion_target = doot.locs.image_temp.with_suffix(ext)
        convert_cmd = self.make_cmd(["convert", str(filepath),
                                    *conversion_args,
                                    str(conversion_target)])
        convert_cmd.execute()

        if doot.locs.image_temp.stat().st_size < 5000000:
            return { 'image': doot.locs.image_temp }

        return False




##-- imports
from __future__ import annotations

import datetime
import json
import logging as root_logger
import pathlib as pl
import uuid
from collections import defaultdict
from shutil import copyfile, rmtree
from typing import (Any, Callable, ClassVar, Dict, Generic, Iterable, Iterator,
                    List, Mapping, Match, MutableMapping, Optional, Sequence,
                    Set, Tuple, TypeVar, Union, cast)

##-- end imports

import networkx as nx

logging = root_logger.getLogger(__name__)

REPLY  : Final[str] = 'in_reply_to_status_id_str'
QUOTE  : Final[str] = 'quoted_status_id_str'
ID_STR : Final[str] = 'id_str'

# TODO could refactor output into template files, ie: jinja.

class TwGraphComponent_M:

    tweet_graph : TwitterGraph

    def tw_construct_component_files(self, tweet_dir:pl.Path, user_dir:pl.Path):
        """
        Create intermediate component files of tweet threads
        creates a tweet_graph of all tweets in tweet_dir,
        then writes individual connected components to output
        returns any missing ids for later download
        """
        assert(self.tweet_graph is not None)
        logging.info("---------- Getting Components")
        components : List[Set[str]]
        components, rest  = self.tweet_graph.components()
        logging.info("Creating %s component files\n\tfrom: %s\n\tto: %s", len(components), tweet_dir, self.output)
        logging.info("Non-Component Tweets: %s", len(rest))

        # Tweet Components allows lazy writing to component files
        # if enough tweets in one component have been queued
        with LazyComponentWriter(self.output, components) as id_map:
            # create separate component files
            logging.info("---------- Copying Tweets to component files")
            # Load each collection of downloaded tweets
            # Note: these are *not* components
            tweet_files = tweet_dir.glob("*.json")
            for jfile in tweet_files:
                data = json.loads(jfile.read_text(), strict=False)
                id_map.add_tweets(data)

            logging.info("---------- Copying Users to component files")
            user_files = user_dir.glob("*.json")
            for jfile in user_files:
                data = json.loads(jfile.read_text(), strict=False)
                assert(isinstance(data, dict))
                id_map.add_users(data)

        logging.info("Component Files Created")
        if bool(id_map.missing):
            logging.info("Missing: (%s) %s", len(id_map.missing), id_map.missing)

        return { "missing" : list(id_map.missing) }

class TwThreadBuild_M:

    def tw_construct_thread(self, fpath:pl.Path):
        """ dfs on components to get order  """
        # Create final orgs, grouped by head user
        logging.info("---------- Constructing Summary for: %s", fpath)
        # read fpath
        data = json.loads(fpath.read_text(), strict=False)
        tweet_lookup = { x['id_str']: x for x in data['tweets'] }
        user_lookup  = { x['id_str']: x for x in data['users'] }
        assert(bool(data))

        thread_graph         = TwitterGraph()
        for tweet in data['tweets']:
            thread_graph.add_tweet(tweet, str(fpath))

        roots  = thread_graph.roots()
        chains, extra = thread_graph.reply_chains(roots)
        if len(chains) > 1:
            logging.warning(f"Component with more than one chain: %s", fpath)
        quotes = thread_graph.get_quotes()
        assert(bool(chains))

        main_thread = max(chains, key=lambda x: len(x))
        main_index  = chains.index(main_thread)
        sub_threads = [x for x in (chains[:main_index] + chains[main_index+1:]) if bool(x)]
        sub_threads += extra

        base_user   = self._calculate_base_user(main_thread, tweet_lookup, user_lookup) or fpath.stem

        thread_obj = TwitterThreadObj(main_thread, sub_threads, quotes, str(fpath), base_user)

        # Then write the thread description out
        # uses the same name as its component
        out_file = self.output / fpath.name.replace("component", "thread")
        out_file.write_text(json.dumps(thread_obj.dump(), indent=4))

    def _calculate_base_user(self, main_thread, tweet_lookup:dict, user_lookup:dict):
        user_counts = defaultdict(lambda: 0)
        try:
            for tweet_id in main_thread:
                tweet   = tweet_lookup.get(tweet_id, {})
                user_id = tweet.get("user", {}).get("id_str", None)
                user    = user_lookup.get(user_id, {}).get("screen_name", None)
                if user:
                    user_counts[user] += 1
            return max(user_counts, key=lambda x: user_counts[x])
        except Exception as err:
            return None

class TwThreadWriting_M:

    def tw_construct_org_files(self, fpath):
        logging.info("Constructing org file from: %s \n\tto: %s", fpath, self.output)
        thread = json.loads(fpath.read_text())
        data   = json.loads(pl.Path(thread['component']).read_text())
        tweets = data['tweets']
        users  = data['users']
        tags   = self.todos.tags_mapping

        thread_writer = OrgThreadWriter.build(thread, tweets, users, tags)
        thread_str = str(thread_writer)

        base_dir    = self.output / "org"
        count       = 0
        target_file =  f"{thread_writer.user}_thread_{count}.org"
        actual_file = base_dir / target_file
        while actual_file.exists():
            count += 1
            target_file =  f"{thread_writer.user}_thread_{count}.org"
            actual_file = base_dir / target_file

        actual_file.write_text(thread_str)

    def tw_construct_html_files(self, fpath):
        logging.info("Constructing html file from: %s \n\tto: %s", fpath, self.output)
        thread = json.loads(fpath.read_text())
        data   = json.loads(pl.Path(thread['component']).read_text())
        tweets = data['tweets']
        users  = data['users']
        tags   = self.todos.tags_mapping

        thread_writer = HtmlThreadWriter.build(thread, tweets, users, tags)
        thread_str = str(thread_writer)

        base_dir    = self.output / "org"
        count       = 0
        target_file =  f"{thread_writer.user}_thread_{count}.html"
        actual_file = base_dir / target_file
        while actual_file.exists():
            count += 1
            target_file =  f"{thread_writer.user}_thread_{count}.html"
            actual_file = base_dir / target_file

        actual_file.write_text(thread_str)



""" Events: The data representations of points and circles
    in the voronoi calculation
"""
##-- imports
from __future__ import annotations
from dataclasses import InitVar, dataclass, field
from enum import Enum
from typing import (Any, Callable, ClassVar, Dict, Generic, Iterable, Iterator,
                    List, Mapping, Match, MutableMapping, Optional, Sequence,
                    Set, Tuple, TypeVar, Union, cast)

import numpy as np

##-- end imports

CIRCLE_EVENTS = Enum("Circle Event Sides", "LEFT RIGHT")

@dataclass
class VEvent:
    """ The Base Class of events in fortunes algorithm """
    loc    : np.ndarray = field()
    step   : int        = field(default=-1)
    offset : float      = field(default=0)

    def y(self):
        """ Get the vertical position of the event """
        return self.loc[1]

    def __lt__(self, other):
        return (VEvent.offset - self.y()) < (VEvent.offset - other.y())

@dataclass
class SiteEvent(VEvent):
    """ Subclass for representing individual points / cell centres """
    face : Face = field(default=None)

    def __str__(self):
        return "Site Event: Loc: {}".format(self.loc)


@dataclass
class CircleEvent(VEvent):
    """ Subclass for representing the lowest point of a circle,
    calculated from three existing site events """

    #The node that will disappear
    sourceNode : Node   = field(default=None)
    #the breakpoint where it will disappear
    #vertex             == centre of circle, not lowest point
    vertex     : Vertex = field(default=None)
    left       : bool   = field(default=True)
    active     : bool   = field(default=True)

    def __post_init__(self, site_loc, sourceNode, voronoiVertex, left=True, i=None):
        if left and (CIRCLE_EVENTS.RIGHT in sourceNode.data
                     and sourceNode.data[CIRCLE_EVENTS.RIGHT].active):
            exception_text = "Trying to add a circle event to a taken left node: {} : {}"
            raise Exception(exception_text.format(sourceNode, sourceNode.data[CIRCLE_EVENTS.RIGHT]))
        elif not left and (CIRCLE_EVENTS.LEFT in sourceNode.data
                           and sourceNode.data[CIRCLE_EVENTS.LEFT].active):
            exception_text = "Trying to add a circle event to a taken right node: {} : {}"
            raise Exception(exception_text.format(sourceNode, sourceNode.data[CIRCLE_EVENTS.LEFT]))

        if left:
            sourceNode.data[CIRCLE_EVENTS.RIGHT] = self
        else:
            sourceNode.data[CIRCLE_EVENTS.LEFT] = self

    def __str__(self):
        return "Circle Event: {}, Node: {}, Left: {}, Added On Step: {}".format(self.loc,
                                                                                self.source,
                                                                                self.left,
                                                                                self.step)

    def deactivate(self):
        """ Deactivating saves on having to reheapify """
        self.active = False



"""
Utility decorators
"""
#pylint: disable=invalid-sequence-index
##-- imports
from __future__ import annotations

import logging as logmod
from enum import Enum
from functools import wraps
from typing import (TYPE_CHECKING, Any, Callable, ClassVar, Dict, Generic,
                    Iterable, Iterator, List, Mapping, Match, MutableMapping,
                    Optional, ParamSpec, Sequence, Set, Tuple, TypeAlias,
                    TypeVar, Union, cast)

##-- end imports

logging = logmod.getLogger(__name__)
T = TypeVar('T')
P = ParamSpec('P')

def ForceListArgDecorator(f:Callable[..., T]) -> Callable[..., T]:
    """ Force the first arg to be a list """

    @wraps(f)
    def wrapped(self, *the_args, **the_kwargs):
        forced = []
        if isinstance(the_args[0], list):
            forced = the_args
        else:
            forced.append([the_args[0]])
            forced += the_args[1:]

        return f(self, *forced, **the_kwargs)

    return wrapped

def cache(f:Callable[..., T]) -> Callable[..., T]:
    cache_key : str = f"{f.__name__}.__cached_val"

    @wraps(f)
    def wrapped(self, *args, **kwargs):
        if hasattr(self, cache_key): #type:ignore
            return cast(T, getattr(self, cache_key)) #type:ignore

        object.__setattr__(self, cache_key, f(self, *args, **kwargs)) #type:ignore
        return getattr(self, cache_key) #type:ignore

    wrapped.__name__ = f"Cached({f.__name__})"
    return wrapped #type:ignore

def invalidate_cache(f:Callable[..., T]) -> Callable[..., T]:
    cache_key = f"{f.__name__}__cached_val"

    @wraps(f)
    def wrapped(self, *args, **kwargs):
        if hasattr(self, cache_key): #type:ignore
            object.__setattr__(self, cache_key, None) #type:ignore

        return f(self, *args, **kwargs)

    wrapped.__name__ = f"CacheInvalidatator({f.__name__})"
    return wrapped

def HandleSignal(sig : str) -> Callable[..., Any]:
    """
    Utility to easily add a signal to classes for use in the handler system
    """

    def __add_sig(sig, method):
        """ used to add 'signal' as an argument to a class' init method """

        @wraps(method)
        def wrapper(self, *args, **kwargs):
            if 'signal' not in kwargs:
                kwargs['signal'] = sig #type:ignore
            return method(self, *args, **kwargs)

        return wrapper

    def wrapper(cls):
        cls.__init__ = __add_sig(sig, cls.__init__) #type:ignore
        return cls

    return wrapper

def factory(f:Callable[..., T]) -> Callable[..., T]:
    """
    Curry a constructor, naming it in a useful way

    """

    @wraps(f)
    def awaiting_arg(first_arg:Any) -> Callable[..., T]:

        @wraps(f)
        def ready_to_apply(*args:Any, **kwargs:Any) -> T:
            return f(first_arg, *args, **kwargs)

        ready_to_apply.__name__ = f"{f.__name__} partial"
        return ready_to_apply

    return awaiting_arg #type:ignore



#!/usr/bin/env python3
"""

"""
##-- imports
from __future__ import annotations

import abc
import datetime
import enum
import functools as ftz
import itertools as itz
import logging as logmod
import pathlib as pl
import re
import time
import types
from copy import deepcopy
from dataclasses import InitVar, dataclass, field
from typing import (TYPE_CHECKING, Any, Callable, ClassVar, Final, Generic,
                    Iterable, Iterator, Mapping, Match, MutableMapping,
                    Protocol, Sequence, Tuple, TypeAlias, TypeGuard, TypeVar,
                    cast, final, overload, runtime_checkable)
from uuid import UUID, uuid1
from weakref import ref

##-- end imports

##-- logging
logging = logmod.getLogger(__name__)
##-- end logging

from functools import partial
from math import atan2, copysign

import cairo_utils as cu
import numpy as np
from cairo_utils.constants import (BEZIER_DATA_LEN, CIRCLE_DATA_LEN, DELTA,
                                   EPSILON, HALFDELTA, LINE_DATA_LEN,
                                   NODE_RECIPROCAL, PI, QUARTERPI,
                                   SAMPLE_DATA_LEN, TOLERANCE, TWOPI,
                                   IntersectEnum)
from numpy import cos, pi, sin
from scipy.interpolate import splev, splprep

def bezier1cp(a_cp_b, t, f=None, p=None):
    """
    Given the start, end, and a control point, create t number of points along that bezier
    t : the number of points to linearly create used to sample along the bezier
    f : a transform function for the sample points prior to calculate bezier
    p : an overriding set of arbitrary sample points for calculate bezier
    """
    #pylint: disable=unused-variable
    assert(isinstance(a_cp_b, np.ndarray))
    if p is not None:
        assert(isinstance(p, np.ndarray))
        sample_points = p
    else:
        sample_points = np.linspace(0, 1, t)
        if f is not None:
            assert(callable(f))
            #f is an easing lookup function
            sample_points = f(t)

    len_samples = len(sample_points)
    quadratic_matrix = np.array([[[1,0,0],[-2,2,0],[1,-2,1]]])
    t_matrix = lambda x: np.array([1, x,  pow(x,2)])

    points = a_cp_b.reshape((-1,3,2))
    results = np.zeros((len(a_cp_b),1,2))
    for t in sample_points:
        t_mul_mat = t_matrix(t) @ quadratic_matrix
        t_lines = t_mul_mat @ points
        results = np.hstack((results, t_lines.reshape(-1,1,2)))
    return results[:,1:,:].reshape((-1,2))

def bezier2cp(a_cpcp_b, n=None, p=None, easing=None):
    """
    Given a start, end, and two control points along the way,
    create n number of points along that bezier
    n : The number of points to sample linearly
    f : the transform function for the linear sampling
    p : arbitrary points to use for sampling instead
    Matrix Multiplication algorithm from:
    https://pomax.github.io/bezierinfo/#decasteljau
    """
    #pylint: disable=too-many-locals
    #pylint: disable=unused-variable
    assert(isinstance(a_cpcp_b, np.ndarray))
    if p is not None:
        assert(isinstance(p, np.ndarray))
        sample_points = p
    elif n is not None:
        sample_points = np.linspace(0, 1, n)
    else:
        raise Exception("Neither arbitrary points or n given")
    if easing is not None:
        assert(callable(easing))
        sample_points = easing(sample_points)

    len_samples = len(sample_points)
    cubic_matrix = np.array([[1,0,0,0],[-3,3,0,0],[3,-6,3,0],[-1,3,-3,1]])
    t_matrix = lambda x: np.array([1, x,  pow(x,2), pow(x,3)])

    points = a_cpcp_b.reshape((-1,4,2))
    results = np.zeros((len(a_cpcp_b),1,2))
    for t in sample_points:
        t_mul_mat = t_matrix(t) @ cubic_matrix
        t_lines = t_mul_mat @ points
        results = np.hstack((results, t_lines.reshape(-1,1,2)))
    return results[:,1:,:].reshape((-1,2))


def is_point_on_line(p, l):
    """ Test to see if a point is on a line """
    assert(isinstance(p, np.ndarray))
    assert(isinstance(l, np.ndarray))
    points = p.reshape((-1, 2))
    the_lines = l.reshape((-1, 2, 2))
    l_mins = the_lines.min(axis=1)
    l_maxs = the_lines.max(axis=1)

    in_bounds_xs = l_mins[:, 0] <= points[:, 0] <= l_maxs[:, 0]
    in_bounds_ys = l_mins[:, 1] <= points[:, 1] <= l_maxs[:, 1]

    if np.allclose((the_lines[:, 0, 0] - the_lines[:, 1, 0]), 0):
        return in_bounds_ys and in_bounds_xs
    slopes = (the_lines[:, 0, 1] - the_lines[:, 1, 1]) / (the_lines[:, 0, 0] - the_lines[:, 1, 0])
    y_intersects = - slopes * the_lines[:, 0, 0] + the_lines[:, 0, 1]
    line_ys = slopes * points[:, 0] + y_intersects
    return np.allclose(line_ys, points[0, 1]) and in_bounds_ys and in_bounds_xs

def make_horizontal_lines(n=1, random=None):
    """ Utility to Describe a horizontal line as a vector of start and end points  """
    if random is None:
        random = np.random.random
    x = random((n, 2)).sort()
    y = random(n).reshape((-1, 1))
    return np.column_stack((x[:, 0], y, x[:, 1], y))

def make_vertical_lines(n=1, random=None):
    """ utility Describe a vertical line as a vector of start and end points """
    if random is None:
        random =np.random.random
    x = random(n).reshape((-1, 1))
    y = random((n, 2)).sort()
    return np.column_stack((x, y[:, 0], x, y[:, 1]))

def sample_along_lines(xys, n, easing=None, override=None):
    """ For a set of lines, sample along them n times,
    with easings distribution. can be hard overriden with override
    """
    t = np.linspace(0,1,n)
    if override is not None:
        t = override
    if easing is not None:
        t = easing(t)
    s_points = t
    s_invert = (1 - s_points)
    num_points = s_points.shape[0]
    fade = np.vstack((s_invert,s_points))
    lines = xys.reshape((-1, 2, 2))
    xs = lines[:, :, 0]
    ys = lines[:, :, 1]
    xsr = xs.repeat(num_points).reshape((-1, 2, num_points))
    ysr = ys.repeat(num_points).reshape((-1, 2, num_points))
    xsrf = xsr * fade
    ysrf = ysr * fade
    s_xs = xsrf.sum(axis=1)
    s_ys = ysrf.sum(axis=1)
    paired = np.hstack((s_xs, s_ys))
    reshaped = paired.reshape((-1, num_points, 2), order='F').reshape((-1,2))
    return reshaped


def create_line(xys, n, easing=None):
    """ Given a start and end, create t number of points along that line """
    line = sample_along_lines(xys.reshape((-1,LINE_DATA_LEN)), n, easing=easing)
    return line

def get_bisector(p1, p2, r=False):
    """ With a normalised line, rotate 90 degrees,
    r=True : to the right
    r=False : to the left
    """
    n = get_unit_vector(p1, p2)
    if r:
        n_prime = n.dot([[0, -1],
                         [1, 0]])
    else:
        n_prime = n.dot([[0, 1],
                         [-1, 0]])
    return n_prime



def intersect(line_1, line_2, tolerance=TOLERANCE):
    """ Get the intersection points of two line segments
    see: http://ericleong.me/research/circle-line/
    so line_1:(start, end), line_2:(start, end)
    returns np.array([x, y]) of intersection or None
    """
    #pylint: disable=too-many-locals
    assert(isinstance(line_1, np.ndarray))
    assert(isinstance(line_2, np.ndarray))
    assert(line_1.shape == (2, 2))
    assert(line_2.shape == (2, 2))
    #The points
    p0 = line_1[0]
    p1 = line_1[1]
    p2 = line_2[0]
    p3 = line_2[1]

    a1 = p1[1] - p0[1]
    b1 = p0[0] - p1[0]
    c1b = a1 * p0[0] + b1 * p0[1]

    a2 = p3[1] - p2[1]
    b2 = p2[0] - p3[0]
    c2b = a2*p2[0] + b2*p2[1]

    detb = a1 * b2 - a2 * b1
    if detb == 0:
        return None

    xb = ((c1b * b2) - (b1 * c2b)) / detb
    yb = ((a1 * c2b) - (c1b * a2)) / detb
    xyb = np.array([xb, yb])

    l1mins = np.min((p0, p1), axis=0) - tolerance
    l2mins = np.min((p2, p3), axis=0) - tolerance
    l1maxs = np.max((p0, p1), axis=0) + tolerance
    l2maxs = np.max((p2, p3), axis=0) + tolerance

    if (l1mins <= xyb).all() and (l2mins <= xyb).all() and \
       (xyb <= l1maxs).all() and (xyb <= l2maxs).all():
        return xyb
    return None


def extend_line(p1, p2, m=1, from_start=True):
    """ Extend a line by m units
    Returns the new end points only
    """
    n = get_unit_vector(p1, p2)
    if from_start:
        el = p1 + (n * m)
    else:
        el = p2 + (n * m)
    return el



#!/usr/bin/env python3
"""


See EOF for license/metadata/notes as applicable
"""

##-- builtin imports
from __future__ import annotations

# import abc
import datetime
import enum
import functools as ftz
import itertools as itz
import logging as logmod
import pathlib as pl
import re
import time
import types
import weakref
# from copy import deepcopy
# from dataclasses import InitVar, dataclass, field
from typing import (TYPE_CHECKING, Any, Callable, ClassVar, Final, Generic,
                    Iterable, Iterator, Mapping, Match, MutableMapping,
                    Protocol, Sequence, Tuple, TypeAlias, TypeGuard, TypeVar,
                    cast, final, overload, runtime_checkable, Generator)
from uuid import UUID, uuid1

##-- end builtin imports

##-- lib imports
import more_itertools as mitz
##-- end lib imports

##-- logging
logging = logmod.getLogger(__name__)
##-- end logging

from jgdv.files.tex.base import TexStatement_i, TexEnvironment_i

class TexFigure(TexEnvironment_i):
    pass

class TexEquation(TexEnvironment_i):
    pass

class TexProof(TexEnvironment_i):
    pass

class TexVerbatim(TexEnvironment_i):
    pass

class TexItemList(TexEnvironment_i):
    pass

class TexTikz(TexEnvironment_i):
    pass

class TexGantt(TexTiz):
    pass


class TexFont(TexEnvironment_i):
    pass

class TexMath(TexEnvironment_i):
    pass



##-- imports
from __future__ import annotations
import logging as logmod
##-- end imports

##-- logging
logging = logmod.getLogger(__name__)
##-- end logging

printer = logmod.getLogger("doot._printer")

import sys
import sh
import jgdv

BACKGROUND = DootKey.make("background")
UPDATE     = DootKey.make("update_")
NOTTY      = DootKey.make("notty")
ENV        = DootKey.make("shenv_")

class JGDVShellBake:

    @DootKey.kwrap.args
    @DootKey.kwrap.types("in_", hint={"on_fail":None, "type_":sh.Command|bool|None})
    @DootKey.kwrap.types("env", hint={"on_fail":sh, "type_":sh.Command|bool|None})
    @DootKey.kwrap.redirects("update_")
    def __call__(self, spec, state, args, _in, env, _update):
        if not env:
            env = sh
        try:
            cmd                     = getattr(env, DootKey.make(args[0], explicit=True).expand(spec, state))
            keys                    = [DootKey.make(x, explicit=True) for x in args[1:]]
            expanded                = [x.expand(spec, state, locs=doot.locs) for x in keys]
            match _in:
                case False | None:
                    baked = cmd.bake(*expanded, _return_cmd=True, _tty_out=False)
                case sh.Command():
                    baked = cmd.bake(*expanded, _in=_in(), _return_cmd=True, _tty_out=False)
                case _:
                    raise DootTaskError("Bad pre-command for shell baking", _in)

            return { _update : baked }
        except sh.CommandNotFound as err:
            printer.error("Shell Commmand '%s' Not Action: %s", err.args[0], args)
        except sh.ErrorReturnCode as err:
            printer.error("Shell Command '%s' exited with code: %s", err.full_cmd, err.exit_code)
            if bool(err.stdout):
                printer.error("%s", err.stdout.decode())

            printer.info("")
            if bool(err.stderr):
                printer.error("%s", err.stderr.decode())

        return False

class JGDVShellBakedRun:

    @DootKey.kwrap.types("in_", hint={"on_fail":None, "type_":sh.Command|None})
    @DootKey.kwrap.redirects("update_")
    def __call__(self, spec, state, _in, _update):
        try:
            result = _in()
            return { _update : result }
        except sh.CommandNotFound as err:
            printer.error("Shell Commmand '%s' Not Action: %s", err.args[0], args)
        except sh.ErrorReturnCode as err:
            printer.error("Shell Command '%s' exited with code: %s", err.full_cmd, err.exit_code)
            if bool(err.stdout):
                printer.error("%s", err.stdout.decode())

            printer.info("")
            if bool(err.stderr):
                printer.error("%s", err.stderr.decode())

        return False

class JGDVShellAction(Action_p):
    """
    For actions in subshells.
    all other arguments are passed directly to the program, using `sh`

    can use a pre-baked sh passed into what "shenv_" points to
    """

    @DootKey.kwrap.args
    @DootKey.kwrap.types("background", "notty", hint={"type_":bool, "on_fail":False})
    @DootKey.kwrap.types("env", hint={"on_fail":sh, "type_":sh.Command|None})
    @DootKey.kwrap.redirects("update_")
    def __call__(self, spec, state, args, background, notty, env, _update) -> dict|bool|None:
        result     = None
        try:
            # Build the command by getting it from env, :
            cmd                     = getattr(env, DootKey.make(args[0], explicit=True).expand(spec, state))
            keys                    = [DootKey.make(x, explicit=True) for x in args[1:]]
            expanded                = [x.expand(spec, state, locs=doot.locs) for x in keys]
            result                  = cmd(*expanded, _return_cmd=True, _bg=background, _tty_out=not notty)
            assert(result.exit_code == 0)

            printer.debug("(%s) Shell Cmd: %s, Args: %s, Result:", result.exit_code, args[0], args[1:])
            if not _update:
                printer.info("%s", result, extra={"colour":"reset"})
                return True

            return { _update : result.stdout }

        except sh.CommandNotFound as err:
            printer.error("Shell Commmand '%s' Not Action: %s", err.args[0], args)
        except sh.ErrorReturnCode as err:
            printer.error("Shell Command '%s' exited with code: %s", err.full_cmd, err.exit_code)
            if bool(err.stdout):
                printer.error("%s", err.stdout.decode())

            printer.info("")
            if bool(err.stderr):
                printer.error("%s", err.stderr.decode())

        return False

class JGDVInteractiveAction(Action_p):
    """
      An interactive command, which uses the self.interact method as a callback for sh.
    """
    _toml_args = ["background"]
    aggregated = ""
    prompt     = ">>> "
    cont       = "... "

    def __call__(self, spec, state:dict) -> dict|bool|None:
        try:
            self.prompt = spec.kwargs.on_fail(JGDVInteractiveAction.prompt, str).prompt()
            self.cont   = spec.kwargs.on_fail(JGDVInteractiveAction.cont, str).cont()

            cmd      = getattr(sh, spec.args[0])
            args     = spec.args[1:]
            keys     = [DootKey.make(x, explicit=True) for x in args]
            expanded = [x.expand(spec, state, locs=doot.locs) for x in keys]
            result   = cmd(*expanded, _return_cmd=True, _bg=spec.kwargs.on_fail(False, bool).background(), _out=self.interact, _out_bufsize=0, _tty_in=True, _unify_ttys=True)
            assert(result.exit_code == 0)
            printer.debug("(%s) Shell Cmd: %s, Args: %s, Result:", result.exit_code, spec.args[0], spec.args[1:])
            printer.info("%s", result, extra={"colour":"reset"})
            return True
        except sh.CommandNotFound as err:
            printer.error("Shell Commmand '%s' Not Action: %s", err.args[0], spec.args)
        except sh.ErrorReturnCode:
            printer.error("Shell Command '%s' exited with code: %s for args: %s", spec[0], result.exit_code, spec.args)

        return False

    def interact(self, char, stdin):
        # TODO possibly add a custom interupt handler
        self.aggregated += str(char)
        if self.aggregated.endswith("\n"):
            printer.info(self.aggregated.strip())
            self.aggregated = ""

        if self.aggregated.startswith(self.prompt) :
            prompt = self.aggregated[:] + ": "
            self.aggregated = ""
            stdin.put(input(prompt) + "\n")
        elif self.aggregated.startswith(self.cont):
            self.aggregated = ""
            val = input(self.cont)
            if bool(val):
                stdin.put("    " + input(self.cont) + "\n")
            else:
                stdin.put(input(self.cont) + "\n")



"""


"""

from jgdv.cli.repl.repl_commander import REPLCommander

register         = REPLCommander.register
register_class   = REPLCommander.register_class
register_default = REPLCommander.register_default



#https://docs.python.org/3/library/unittest.html
# https://docs.python.org/3/library/unittest.mock.html

from os.path import splitext, split

import unittest
import unittest.mock as mock

import logging


class FileProcessingTests(unittest.TestCase):

    @classmethod
    def setUpClass(cls):
        logmod.getLogger('').setLevel(logmod.WARNING)
        LOG_FILE_NAME = "log.{}".format(splitext(split(__file__)[1])[0])

        file_h = logmod.FileHandler(LOG_FILE_NAME, mode='w')
        file_h.setLevel(logmod.DEBUG)

        console = logmod.StreamHandler()
        console.setLevel(logmod.WARNING)

        logging = logmod.getLogger(__name__)
        logging.setLevel(logmod.DEBUG)
        logging.addHandler(console)
        logging.addHandler(file_h)


    def setUp(self):
        return 1

    def tearDown(self):
        return 1

    #----------
    # use testcase snippet
    # mock.Mock / MagicMock
    # create_autospec
    # @patch(' ') / with patch.object(...)



#!/usr/bin/env python3
"""

See EOF for license/metadata/notes as applicable
"""

##-- builtin imports
from __future__ import annotations

# import abc
import datetime
import enum
import functools as ftz
import itertools as itz
import logging as logmod
import pathlib as pl
import re
import time
import types
import weakref
# from copy import deepcopy
# from dataclasses import InitVar, dataclass, field
from typing import (TYPE_CHECKING, Any, Callable, ClassVar, Final, Generic,
                    Iterable, Iterator, Mapping, Match, MutableMapping,
                    Protocol, Sequence, Tuple, TypeAlias, TypeGuard, TypeVar,
                    cast, final, overload, runtime_checkable, Generator)
from uuid import UUID, uuid1

##-- end builtin imports

##-- lib imports
import more_itertools as mitz
##-- end lib imports

##-- logging
logging = logmod.getLogger(__name__)
printer = logmod.getLogger("doot._printer")
##-- end logging

import base64

from selenium.webdriver import FirefoxOptions, FirefoxService, Firefox
from selenium.webdriver.common.print_page_options import PrintOptions

FF_DRIVER     = "__$ff_driver"
READER_PREFIX = "about:reader?url="

def setup_firefox(spec, state):
    """ Setups a selenium driven, headless firefox to print to pdf """
    printer.info("Setting up headless Firefox")
    options = FirefoxOptions()
    # options.add_argument("--start-maximized")
    options.add_argument("--headless")
    # options.binary_location = "/usr/bin/firefox"
    # options.binary_location = "/snap/bin/geckodriver"
    options.set_preference("print.always_print_silent", True)
    options.set_preference("print.printer_Mozilla_Save_to_PDF.print_to_file", True)
    options.set_preference("print_printer", "Mozilla Save to PDF")
    options.set_preference("print.printer_Mozilla_Save_to_PDF.use_simplify_page", True)
    options.set_preference("print.printer_Mozilla_Save_to_PDF.print_page_delay", 50)
    service = FirefoxService(executable_path="/snap/bin/geckodriver")
    driver = Firefox(options=options, service=service)
    return { FF_DRIVER : driver }

@DootKey.kwrap.expands("url")
@DootKey.kwrap.paths("to")
@DootKey.kwrap.types(FF_DRIVER)
def save_pdf(spec, state, url, _to, _driver):
    """ prints a url to a pdf file using selenium """
    printer.info("Saving: %s", url)
    print_ops = PrintOptions()
    print_ops.page_range = "all"

    driver.get(READER_PREFIX + url)
    time.sleep(2)
    pdf       = _driver.print_page(print_options=print_ops)
    pdf_bytes = base64.b64decode(pdf)

    with open(_to, "wb") as f:
        f.write(pdf_bytes)

@DootKey.kwrap.types(FF_DRIVER)
def close_firefox(spec, state, _driver):
    printer.info("Closing Firefox")
    _driver.quit()



#!/usr/bin/env python3
"""


See EOF for license/metadata/notes as applicable
"""

##-- builtin imports
from __future__ import annotations

# import abc
import datetime
import enum
import functools as ftz
import itertools as itz
import logging as logmod
import pathlib as pl
import re
import time
import types
import weakref
# from copy import deepcopy
# from dataclasses import InitVar, dataclass, field
from typing import (TYPE_CHECKING, Any, Callable, ClassVar, Final, Generic,
                    Iterable, Iterator, Mapping, Match, MutableMapping,
                    Protocol, Sequence, Tuple, TypeAlias, TypeGuard, TypeVar,
                    cast, final, overload, runtime_checkable, Generator)
from uuid import UUID, uuid1

##-- end builtin imports

##-- lib imports
import more_itertools as mitz
##-- end lib imports

##-- logging
logging = logmod.getLogger(__name__)
##-- end logging

@dataclass
class DCELDrawSettings:
    text     : bool = field(default=False)
    faces    : bool = field(default=True)
    edges    : bool = field(default=False)
    _vertices : bool = field(default=False)

    face_colour : Colour = field(default=FACE_COLOUR)
    edge_colour : Colour = field(default=EDGE_COLOUR)
    vert_colour : Colour = field(default=VERT_COLOUR)
    background  : Colour = field(default=BACKGROUND_COLOUR)

    edge_width  : float  = field()



"""

"""
##-- imports
from __future__ import annotations

import abc
import importlib
import logging as logmod
import re
import tracemalloc
from collections import defaultdict
from dataclasses import InitVar, dataclass, field
from datetime import datetime
from enum import Enum
from os.path import abspath, exists, expanduser, split, splitext
from typing import (TYPE_CHECKING, Any, Callable, ClassVar, Final, Generic,
                    Iterable, Iterator, Mapping, Match, MutableMapping,
                    Protocol, Sequence, Tuple, TypeAlias, TypeGuard, TypeVar,
                    cast, final, overload, runtime_checkable)

##-- end imports

import pyparsing as pp

logging = logmod.getLogger(__name__)
config  = acab.config

snap_a  = None

def display_memory(self):
    """
    A Post Command Hook for ReplCommander, to print the current and peak memory
    usage.
    """
    global snap_a
    traced  = tracemalloc.get_traced_memory()
    print("Memory: Current: {:<5}, Peak: {:<5}".format(human(traced[0]), human(traced[1])))

@register
def do_memdiff(self, line):
    """
    Print the top 20 differences between the last snapshot and now
    """
    current = tracemalloc.take_snapshot().filter_traces(sh_filter)
    diff    = current.compare_to(snap_a, 'lineno')
    for x in diff[:20]:
        print_diff(x)

@register
def do_memstat(self, line):
    """
    Print the top 20 File based memory statistics
    """
    if not tracemalloc.is_tracing():
        print("Call memstat after starting memory tracing with `memory`")
        return
    current = tracemalloc.take_snapshot().filter_traces(sh_filter)
    print("\nBy Filename --------------------")
    for stat in current.statistics("filename")[:20]:
        print_stat_file(stat)

    print()

@register
def do_snapshot(self, line):
    """
    Take a Memory Snapshot for comparison
    """
    global snap_a, snap_b
    if not tracemalloc.is_tracing():
        print("Must start memory tracing to take a snapshot")
        print("Call `memory`")
        return

    snap_a = tracemalloc.take_snapshot().filter_traces(sh_filter)

@register
def do_memory(self, line):
    """
    Toggle Memory Tracing
    """
    limit = 50
    if display_memory.__name__ in self.state.post_cmds:
        print("Disabling Memory Trace")
        del self.state.post_cmds[display_memory.__name__]
        tracemalloc.stop()
        return

    print(f"Adding Memory Trace with limit {limit}")
    tracemalloc.start(limit)
    do_snapshot(self, "")
    self.state.post_cmds[display_memory.__name__] = display_memory



#!/usr/bin/env python3
"""

"""
##-- imports
from __future__ import annotations

import logging as logmod
import unittest
import warnings
import pathlib as pl
from typing import (Any, Callable, ClassVar, Generic, Iterable, Iterator,
                    Mapping, Match, MutableMapping, Sequence, Tuple, TypeAlias,
                    TypeVar, cast)
##-- end imports
logging = logmod.root

import pytest

@pytest.mark.parametrize("ctor", [DootTracker])
class TestTracker:

    def test_initial(self, ctor):
        tracker = ctor()
        assert(tracker is not None)

    def test_add_task(self, ctor, mocker):
        mock_task = mock_gen.mock_task("task1")
        tracker = ctor()
        tracker.add_task(mock_task)

        assert("task1" in tracker.tasks)
        assert(tracker.task_graph.nodes['task1']['state'] == tracker.state_e.DEFINED)

    def test_duplicate_add_fail(self, ctor, mocker):
        """ dont add a duplicately named task, or its dependencies """
        task1 = mock_gen.mock_task("task1")
        task2 = mock_gen.mock_task("task1")

        tracker = ctor()
        tracker.add_task(task1)
        with pytest.raises(doot.errors.DootTaskTrackingError):
            tracker.add_task(task2)

        assert("task1" in tracker.tasks)

    def test_duplicate_add(self, ctor, mocker):
        task1 = mock_gen.mock_task("task1")
        task2 = mock_gen.mock_task("task1")

        tracker = ctor(shadowing=True)
        tracker.add_task(task1)
        tracker.add_task(task2)

        assert("task1" in tracker.tasks)

    def test_contains_defined(self, ctor, mocker):
        mock_task = mock_gen.mock_task("test_task")
        mock_task.depends_on = ["example", "blah"]

        tracker = ctor()
        tracker.add_task(mock_task)
        # defined Task is contained
        assert("test_task" in tracker)


    def test_warn_on_undefined(self, ctor, mocker, caplog):
        """ create a task with undefined dependencies, it should just warn not error """
        tracker = ctor()
        for task in mock_gen.task_network({"task1"     : [["subtask", "subtask2"], []]}):
            tracker.add_task(task)

        assert(tracker.next_for("task1").name is "task1")
        assert(bool([x for x in caplog.records if x.levelname == "WARNING"]))
        assert("Tried to Schedule a Declared but Undefined Task: subtask" in caplog.messages)
        assert("Tried to Schedule a Declared but Undefined Task: subtask2" in caplog.messages)

    def test_not_contains_declared(self, ctor, mocker):
        mock_task = mock_gen.mock_task("test_task", pre=["example", "blah"])

        tracker = ctor()
        tracker.add_task(mock_task)
        assert("example" not in tracker)
        assert("blah" not in tracker)

    def test_task_prior_registration(self, ctor, mocker):
        mock_task = mock_gen.mock_task("test_task")
        mock_task.depends_on.append("example")
        mock_task.depends_on.append("blah")

        tracker = ctor()
        tracker.add_task(mock_task)
        assert(tracker.task_graph.nodes['example']['state'] == tracker.state_e.DECLARED)
        assert(tracker.task_graph.nodes['blah']['state'] == tracker.state_e.DECLARED)
        assert("example" in tracker.task_graph)
        assert("blah" in tracker.task_graph)

    def test_task_post_registration(self, ctor, mocker):
        mock_task = mock_gen.mock_task("test_task")
        mock_task.required_for.append("example")
        mock_task.required_for.append("blah")

        tracker = ctor()
        tracker.add_task(mock_task)
        assert(tracker.task_graph.nodes['example']['state'] == tracker.state_e.DECLARED)
        assert(tracker.task_graph.nodes['blah']['state'] == tracker.state_e.DECLARED)
        assert("example" in tracker.task_graph)
        assert("blah" in tracker.task_graph)

    def test_declared_set(self, ctor, mocker):
        mock_task = mock_gen.mock_task("test_task")
        mock_task.depends_on   += ["subtask", "sub2"]
        mock_task.required_for += ["example", "blah"]

        tracker = ctor()
        tracker.add_task(mock_task)
        declared = tracker.declared_set()
        assert(declared == {"__root", "test_task", "subtask","sub2", "example", "blah"})

    def test_defined_set(self, ctor, mocker):
        mock_task = mock_gen.mock_task("test_task")

        tracker = ctor()
        tracker.add_task(mock_task)

        defined = tracker.defined_set()
        assert(defined == {"test_task"})

    def test_task_order(self, ctor, mocker):
        tasks = mock_gen.task_network({
            "task1"   : [["subtask", "subtask2"], []],
            "subtask" : [["subsub"], []],
            "subtask2": [["subsub"], []],
         })

        tracker  = ctor()
        for task in tasks:
            tracker.add_task(task)

        next_task = tracker.next_for("task1")
        assert(next_task.name in {"subtask", "subtask2"})
        tracker.update_state(next_task, tracker.state_e.SUCCESS)
        next_task_2 = tracker.next_for()
        assert(next_task_2.name in {"subtask", "subtask2"} - {next_task.name})
        tracker.update_state(next_task_2, tracker.state_e.SUCCESS)
        next_task_3 = tracker.next_for()
        assert(next_task_3.name in "task1")

    def test_task_iter(self, ctor, mocker):
        tasks = mock_gen.task_network({
            "task1"   : [["subtask", "subtask2", "subtask3"], []],
            "subtask" : [["subsub"], []],
            "subtask2": [["subsub"], []],
            "subtask3": [["subsub"], []],
         })

        tasks[0].priority = 0

        tracker             = ctor()
        for task in tasks:
            tracker.add_task(task)

        result_tasks = []
        tracker.queue_task("task1")
        for x in tracker:
            if x:
                result_tasks.append(x.name)
                tracker.update_state(x.name, tracker.state_e.SUCCESS)

        assert(len(result_tasks) == 4)

    def test_task_iter_state_changed(self, ctor, mocker):
        tracker  = ctor()
        for task in mock_gen.task_network({"task1"   : [["subtask", "subtask2", "subtask3"], []],
                                           "subtask" : [["subsub"], []],
                                           "subtask2": [["subsub"], []],
                                           "subtask3": [["subsub"], []],
                                          }):
            tracker.add_task(task)

        tracker.update_state("subtask2", tracker.state_e.SUCCESS)
        tasks = []
        tracker.queue_task("task1")
        for x in tracker:
            if x:
                tasks.append(x.name)
                tracker.update_state(x.name, tracker.state_e.SUCCESS)

        assert("subtask2" not in tasks)
        assert(len(tasks) == 3)

    def test_task_failure(self, ctor, mocker, caplog):
        tracker = ctor()
        for task in mock_gen.task_network({"task1"   : [["subtask"], []],
                                           "subtask" : [["subsub"], []],
                                          }):
            tracker.add_task(task)

        result = tracker.next_for("task1")
        assert(result.name == "subtask")
        tracker.update_state(result, tracker.state_e.SUCCESS)
        assert(tracker.next_for().name == "task1")
        assert("Tried to Schedule a Declared but Undefined Task: subsub" in caplog.messages)


    def test_post_task_order(self, ctor, mocker):
        tracker = ctor()
        for task in mock_gen.task_network({"task1"     : [["subtask", "subtask2"], []],
                                           "subtask"   : [["subsub"], ["sidesuper"]],
                                           "subtask2"  : [[], []],
                                           "subsub"    : [[], []],
                                           "sidesuper" : [[], []],

                                          }):
            tracker.add_task(task)

        next_task = tracker.next_for("task1")
        assert(next_task.name == "subtask2")
        tracker.update_state(next_task, tracker.state_e.SUCCESS)

        next_task_2 = tracker.next_for()
        assert(next_task_2.name == "subsub")
        tracker.update_state(next_task_2, tracker.state_e.SUCCESS)

        next_task_3 = tracker.next_for()
        assert(next_task_3.name == "subtask")
        tracker.update_state(next_task_3, tracker.state_e.SUCCESS)

        next_task_4 = tracker.next_for()
        assert(next_task_4.name == "task1" )

    def test_task_exact_artifact_dependency(self, ctor, mocker):
        tracker = ctor()
        for task in mock_gen.task_network({"task1"     : [[pl.Path("test.file")], []],
                                           "subtask"   : [[], [pl.Path("blah.other")]],
                                           "subtask2"  : [[], [pl.Path("test.file")]],

                                          }):
            tracker.add_task(task)

        next_task = tracker.next_for("task1")
        assert(next_task.name == "subtask2")

    def test_task_inexact_artifact_dependency(self, ctor, mocker):
        tracker = ctor()
        for task in mock_gen.task_network({"task1"     : [[pl.Path("*.file")], []],
                                           "subtask"   : [[], [pl.Path("blah.other")]],
                                           "subtask2"  : [[], [pl.Path("test.file")]],

                                          }):
            tracker.add_task(task)

        next_task = tracker.next_for("task1")
        assert(isinstance(next_task, doot.structs.DootTaskArtifact))

    def test_task_artifact_exists(self, ctor, mocker):
        """
          check that if artifacts exist, tasks that generate them aren't queued
        """
        mocker.patch.object(pl.Path, "exists", return_value=True)
        tracker = ctor()
        for task in mock_gen.task_network({"task1"     : [[pl.Path("*.file")], []],
                                           "subtask"   : [[], [pl.Path("blah.other")]],
                                           "subtask2"  : [[], [pl.Path("test.file")]],

                                          }):
            tracker.add_task(task)


        next_task = tracker.next_for("task1")
        assert(isinstance(next_task, doot.structs.DootTaskArtifact))

    def test_task_artifact_doesnt_exists(self, ctor, mocker):
        mocker.patch.object(pl.Path, "exists", return_value=False)
        tracker = ctor()
        for task in mock_gen.task_network({"task1"     : [[pl.Path("*.file")], []],
                                           "subtask"   : [[], [pl.Path("blah.other")]],
                                           "subtask2"  : [[], [pl.Path("test.file")]],

                                          }):
            tracker.add_task(task)

        next_task = tracker.next_for("task1")
        assert(isinstance(next_task, doot.structs.DootTaskArtifact))

    def test_task_artifact_partial_exists(self, ctor, mocker):

        def temp_exists(self):
            return not "*" in self.stem

        mocker.patch.object(pl.Path, "exists", new=temp_exists)
        tracker = ctor()
        for task in mock_gen.task_network({"task1"     : [[pl.Path("*.file")], []],
                                           "subtask"   : [[], [pl.Path("blah.other")]],
                                           "subtask2"  : [[], [pl.Path("test.file")]],

                                          }):
            tracker.add_task(task)

        next_task = tracker.next_for("task1")
        assert(isinstance(next_task, doot.structs.DootTaskArtifact))



"""

"""
from __future__ import annotations

import logging as logmod
from dataclasses import InitVar, dataclass, field
from typing import (TYPE_CHECKING, Any, Callable, ClassVar, Final, Generic,
                    Iterable, Iterator, Mapping, Match, MutableMapping,
                    Protocol, Sequence, Tuple, TypeAlias, TypeGuard, TypeVar,
                    cast, final, overload, runtime_checkable)

logging = logmod.getLogger(__name__)

@dataclass
class ReplState:
    """ Data used for control of the repl """
    prompt        : str                            =  field(default="Prompt: ")
    prompt_ml     : str                            =  field(default="MLPrompt: ")
    prompt_bkup   : str                            =  field(default="")
    ctxs          : ContextSet_i                   =  field(default=None)
    collect_str   : list[str]                      =  field(default_factory=list)
    echo          : bool                           =  field(default=False)
    in_multi_line : bool                           =  field(default=False)
    indent        : int                            =  field(default=0)
    debug_data    : Any                            =  field(default=None)
    debugger      : Any                            =  field(default=None)
    last_err      : Any                            =  field(default=None)
    # run in alphabetical order:
    post_cmds     : dict[str, Callable[..., None]] =  field(default_factory=dict)



"""
Provides the main RBTree class
"""
from functools import partial
from types import FunctionType
import logging as root_logger

from .operations import rb_tree_delete, rb_tree_fixup
from .comparison_functions import Directions, default_comparison, default_equality
from .node import Node


logging = root_logger.getLogger(__name__)

class RBTree:
    """ A Red-Black Tree Implementation
    Properties of RBTrees:
    1) Every node is Red Or Black
    2) The root is black
    3) Every leaf is Black, leaves are null nodes
    4) If a node is red, it's children are black
    5) All paths from a node to its leaves contain the same number of black nodes
    """

    def __init__(self, cmp_func=None, eq_func=None, cleanup_func=None):
        """ Initialise the rb tree container, ie: the node list """
        #Default Comparison and Equality functions with dummy data ignored
        assert(isinstance(cmp_func, (None, partial, FunctionType)))
        assert(isinstance(eq_func, (None, partial, FunctionType)))

        self.nodes        = []
        self.root         = None
        self.cmp_func     = cmp_func or default_comparison
        self.eq_func      = eq_func or default_equality
        self.cleanup_func = cleanup_func or (lambda x: [])


    #------------------------------
    # def Basic Access
    #------------------------------
    def __len__(self):
        return len(self.nodes)

    def __repr__(self):
        if self.root is None:
            return "RBTree(_)"

        return "RBTree( Len: {})".format(len(self))

    def min(self):
        """ Get the min value of the tree """
        if self.root is None:
            return None
        return self.root.min()

    def max(self):
        """ Get the max value of the tree """
        if self.root is None:
            return None
        return self.root.max()

    def get_chain(self):
        """ Get the sequence of leaf values, from left to right """
        if self.root is None:
            return []
        found = set()
        chain = []
        current = self.root.min()
        while current is not None:
            if current in found:
                raise Exception("This shouldn't happen")
            found.add(current)
            chain.append(current)
            current = current.get_successor()
        return chain

    def get_successor_triple(self, node):
        """ Get the trio of this node and two successors """
        if node is None:
            return None
        a = node
        b = a.get_successor()
        if b is not None:
            c = b.get_successor()
            if c is not None:
                return (a, b, c)
        return None

    def get_predecessor_triple(self, node):
        """ Get the trio of this node and two predecessors """
        if node is None:
            return None
        a = node
        b = a.getPredecessor()
        if b is not None:
            c = b.getPredecessor()
            if c is not None:
                return (c, b, a)
        return None


    #------------------------------
    # def Query
    #------------------------------
    def search(self, value, cmp_func=None, eq_func=None,
               cmp_data=None, closest=False, start=None):
        """ Search the tree for a value """
        logging.debug("Searching for: {}".format(value))

        if cmp_func is None:
            cmp_func = self.cmp_func
        if eq_func is None:
            eq_func = self.eq_func
        if start is None:
            start = self.root
        parent = start
        current = start
        comp = None
        while current is not None and not eq_func(current, value, cmp_data):
            parent = current
            comp = cmp_func(current, value, cmp_data)
            if comp is Directions.LEFT:
                logging.debug("left")
                current = current.left
            else:
                assert(comp is Directions.RIGHT)
                logging.debug("right")
                current = current.right

        if closest and current is None:
            #closest non-exact match found
            logging.debug("Found Closest: {}, {}".format(parent, comp))
            return (parent, comp)
        elif current is None:
            #nothing found
            logging.debug("Found Nothing")
            return (None, None)
        else:
            #exact match found
            logging.debug("Found Exact: {}, {}".format(current, comp))
            return (current, comp)


    #------------------------------
    # def Public Update
    #------------------------------
    def update_values(self, func, func_data):
        """ Call a function on all stored values,
        function signature:  f(value, func_data)
        """
        assert(isinstance(func, (FunctionType, partial)))
        for node in self.nodes:
            func(node.value, func_data)

    def insert(self, *args, data=None, cmp_data=None):
        """ Insert data into the tree """
        nodes = []
        for x in args:
            new_node = self.__insert(x, data=data, cmp_data=cmp_data)
            nodes.append(new_node)
        return nodes

    def delete(self, *args, cleanup_func=None):
        """ Delete a value from the tree """
        if cleanup_func is None:
            cleanup_func = self.cleanup_func
        to_remove = set(args)
        while bool(to_remove):
            target = to_remove.pop()
            logging.debug("Removing Target: {}".format(target))
            assert(isinstance(target, Node))
            if target not in self.nodes:
                continue
            to_remove.update(cleanup_func(target))
            rb_tree_delete(self, target)
            self.nodes.remove(target)

    def delete_value(self, *args, cmp_func=None, eq_func=None, cleanup_func=None,
                     cmp_data=None):
        """ Delete a value from the tree, and rebalance  """
        for val in args:
            node, _ = self.search(val, cmp_func=cmp_func,
                                  eq_func=eq_func, cmp_data=cmp_data)
            if node is not None:
                self.delete(node, cleanup_func=cleanup_func)
            else:
                logging.debug("Not Found: {}".format(val))


    #------------------------------
    # def Private Update
    #------------------------------
    def __insert(self, value, data=None, cmp_data=None):
        """ Insert a value into the tree """
        parent, direction = self.search(value, closest=True, cmp_data=cmp_data)
        if direction is Directions.LEFT:
            return self.insert_predecessor(parent, value, data=data)
        else:
            return self.insert_successor(parent, value, data=data)

    def insert_successor(self, existing_node, new_value, data=None):
        """ Insert a new value as a successor to a specific node """
        assert(existing_node is None or isinstance(existing_node, Node))
        logging.debug("Inserting Successor")
        new_node = Node(new_value, data=data)
        self.nodes.append(new_node)
        if existing_node is None:
            self.root = new_node
        else:
            existing_node.add_right(new_node)
        self.__balance(new_node)
        return new_node

    def insert_predecessor(self, existing_node, new_value, data=None):
        """ Insert a new value as a predecessor to the specific node """
        assert(existing_node is None or isinstance(existing_node, Node))
        logging.debug("Inserting predecessor")
        new_node = Node(new_value, data=data)
        self.nodes.append(new_node)
        if existing_node is None:
            self.root = new_node
        else:
            existing_node.add_left(new_node)
        self.__balance(new_node)
        return new_node

    def __balance(self, node):
        assert(isinstance(node, Node))
        rb_tree_fixup(self, node)


    #------------------------------
    # def Debug
    #------------------------------
    def count_black_height(self, node=None):
        """ Given a node, count all paths and check they have the same black height """
        if node is None:
            if self.root is None:
                return None
            node = self.root
        stack = [node]
        leaves = []
        while bool(stack):
            current = stack.pop()
            if current.left is None and current.right is None:
                leaves.append(current)
            else:
                if current.left is not None:
                    stack.append(current.left)
                if current.right is not None:
                    stack.append(current.right)

        all_heights = [x.getBlackHeight(node) for x in leaves]
        return all_heights



#!/usr/bin/env python3
"""


See EOF for license/metadata/notes as applicable
"""

##-- builtin imports
from __future__ import annotations

# import abc
import datetime
import enum
import functools as ftz
import itertools as itz
import logging as logmod
import pathlib as pl
import re
import time
import types
import weakref
# from copy import deepcopy
# from dataclasses import InitVar, dataclass, field
from typing import (TYPE_CHECKING, Any, Callable, ClassVar, Final, Generic,
                    Iterable, Iterator, Mapping, Match, MutableMapping,
                    Protocol, Sequence, Tuple, TypeAlias, TypeGuard, TypeVar,
                    cast, final, overload, runtime_checkable, Generator)
from uuid import UUID, uuid1

##-- end builtin imports

##-- lib imports
import more_itertools as mitz
##-- end lib imports

##-- logging
logging = logmod.getLogger(__name__)
##-- end logging

import networkx as nx

@dataclass
class LazyComponentWriter:
    """
    An Accessor for appending data to files easily.
    comp_dir is the directory to write to
    components is the sets of tweet id's forming components

    component files are json dicts: {"tweets: [], "users": [] }

    """
    comp_dir      : pl.Path                         = field()
    components    : List[Set[str]]                  = field()

    tweet_mapping : Dict[str, set[ComponentWriter]] = field(default_factory=lambda: defaultdict(lambda: set()))
    user_mapping  : Dict[str, set[ComponentWriter]] = field(default_factory=lambda: defaultdict(lambda: set()))
    writers       : List[ComponentWriter]           = field(default_factory=list)
    missing       : Set[str]                        = field(default_factory=set)

    def __post_init__(self):
        # For each component
        for comp_set in self.components:
            # Create a writer
            comp_obj = ComponentWriter(self.comp_dir)
            assert(not comp_obj.path.exists())
            # And pair each tweet id with that writer
            self.writers.append(comp_obj)
            for x in comp_set:
                self.tweet_mapping[x].add(comp_obj)

    def __contains__(self, value):
        return value in self.tweet_mapping

    def finish(self):
        for comp_f in self.writers:
            comp_f.finish()

        self._write_summary()

    def _write_summary(self):
        summary_path = self.comp_dir / "components.summary"
        lines = [x.summary() for x in self.writers]
        summary_path.write_text("\n".join(lines))

    def add_tweets(self, data:list):
        for tweet in data:
            id_s    = tweet.get('id_str', None)
            user_id = tweet.get('user', {}).get('id_str', None)
            if id_s is None:
                continue
            if id_s not in self.tweet_mapping:
                self.missing.add(id_s)
                continue

            for comp_f in self.tweet_mapping[id_s]:
                comp_f.add(tweet)
                # Each tweet maps its user to the writer as well, for the user pass
                self.user_mapping[user_id].add(comp_f)


    def add_users(self, data:dict):
        for user_id, user in data.items():
            if user_id not in self.user_mapping:
                self.missing.add(user_id)
                continue

            for comp_f in self.user_mapping[user_id]:
                comp_f.add(user, data_type="users")

    def __enter__(self):
        return self

    def __exit__(self, atype, value, traceback):
        self.finish()



"""
Math Utilities
TODO: Refactor into submodules
"""
##-- imports
from __future__ import annotations

import logging as root_logger
from functools import partial
from math import atan2, copysign
import numpy as np
from numpy import cos, sin, pi
from scipy.interpolate import splprep, splev

from ..constants import PI, TWOPI, QUARTERPI, EPSILON, TOLERANCE
from ..constants import IntersectEnum, DELTA, HALFDELTA, NODE_RECIPROCAL
from ..constants import SAMPLE_DATA_LEN, LINE_DATA_LEN, BEZIER_DATA_LEN, CIRCLE_DATA_LEN

##-- end imports

from functools import partial
from math import atan2, copysign

import cairo_utils as cu
import numpy as np
from cairo_utils.constants import (BEZIER_DATA_LEN, CIRCLE_DATA_LEN, DELTA,
                                   EPSILON, HALFDELTA, LINE_DATA_LEN,
                                   NODE_RECIPROCAL, PI, QUARTERPI,
                                   SAMPLE_DATA_LEN, TOLERANCE, TWOPI,
                                   IntersectEnum)
from numpy import cos, pi, sin
from scipy.interpolate import splev, splprep

logging = root_logger.getLogger(__name__)

def _interpolate(xy, num_points, smoothing=0.2):
    """ given a set of points, generate values between those points """
    assert(isinstance(xy, np.ndarray))
    assert(len(xy.shape) == 2)
    assert(xy.shape[0] >= 4)
    assert(xy.shape[1] == 2)
    spline_tuple, _ = splprep([xy[:, 0], xy[:, 1]], s=smoothing)
    interpolate_points = np.linspace(0, 1, num_points)
    smoothed_xy = np.column_stack(splev(interpolate_points, spline_tuple))
    return smoothed_xy

def scale(xs, minmax):
    result = minmax[0] + (xs * (minmax[1] - minmax[0]))
    return result

def scale_ndarray(xs, minmaxs):
    mins = minmaxs[:,0].reshape((-1,1))
    ranges = (minmaxs[:,1] - minmaxs[:,0]).reshape((-1,1))
    result = mins + (xs * ranges)
    return result

def check_sign(a, b):
    """ Test whether two numbers have the same sign """
    return copysign(a, b) == a

def get_min_range_pair(p1, p2):
    """ TODO: Can't remember, test this """
    d1 = get_distance(p1, p2)
    fp2 = np.flipud(p2)
    d2 = get_distance(p1, fp2)
    d1_min = d1.min()
    d2_min = d2.min()
    if d1_min < d2_min:
        i = d1.tolist().index(d1_min)
        #get the right xs
        return np.array([p1[i][0], p2[i][0]])
    else:
        i = d2.tolist().index(d2_min)
        return np.array([p1[i][0], fp2[i][0]])

def clamp(n, minn=0, maxn=1):
    """ Clamp a number between min and max,
    could be replaced with np.clip
    """
    return max(min(maxn, n), minn)

def get_ranges(a):
    """ Given pairs, get the ranges of them """
    assert(isinstance(a, np.ndarray))
    assert(a.shape[1] == 2)
    ranges = np.array([a.min(axis=0), a.max(axis=0)])
    return ranges.T

def node_to_position(x, y):
    """ Convert a nodes XY Position to its actual position """
    return [NODE_RECIPROCAL * x, NODE_RECIPROCAL * y]



#!/usr/bin/env python3
"""

"""
##-- imports
from __future__ import annotations

# import abc
# import datetime
import enum
import functools as ftz
import itertools as itz
import logging as logmod
import pathlib as pl
import re
import time
import types
# from copy import deepcopy
# from dataclasses import InitVar, dataclass, field
from typing import (TYPE_CHECKING, Any, Callable, ClassVar, Final, Generic,
                    Iterable, Iterator, Mapping, Match, MutableMapping,
                    Protocol, Sequence, Tuple, TypeAlias, TypeGuard, TypeVar,
                    cast, final, overload, runtime_checkable, Generator, Literal)
# from uuid import UUID, uuid1
# from weakref import ref

# from bs4 import BeautifulSoup
import boltons.queueutils
# import construct as C
# import dirty-equals as deq
# import graphviz
# import matplotlib.pyplot as plt
# import more_itertools as itzplus
import networkx as nx
# import numpy as np
# import pandas
# import pomegranate as pom
# import pony import orm
# import pronouncing
# import pyparsing as pp
# import rich
# import seaborn as sns
# import sklearn
# import stackprinter # stackprinter.set_excepthook(style='darkbg2')
# import sty
# import sympy
# import tomllib
# import toolz
# import tqdm
# import validators
# import z3
# import spacy # nlp = spacy.load("en_core_web_sm")

##-- end imports

##-- logging
logging = logmod.getLogger(__name__)
##-- end logging

from collections import defaultdict
from queue import PriorityQueue
import jgdv
from jgdv.enums.task_state import TaskStateEnum

@jgdv.check_protocol
class JGDVReactorTracker(BaseTracker, TaskTracker_i):
    """
    track dependencies in a networkx digraph,
    successors of a node are its dependencies.
      ie: ROOT -> Task -> Dependency -> SubDependency

    tracks definite and indefinite artifacts as products and dependencies of tasks as well.

    the `task_graph` stores nodes as full names of tasks
    """
    state_e = TaskStateEnum

    def __init__(self, shadowing:bool=False, *, policy=None):
        super().__init__(policy=policy) # self.tasks

    def add_task(self, task:JGDVTaskSpec|TaskBase_i, *, no_root_connection=False) -> None:
        """ add a task description into the tracker, but don't queue it
        connecting it with its dependencies and tasks that depend on it
        """

        task : TaskBase_i = self._prep_task(task)
        assert(isinstance(task, TaskBase_i))
        # Store it
        self.tasks[task.name] = task

        # Insert into dependency graph
        self.task_graph.add_node(task.name, state=self.INITIAL_TASK_STATE, priority=task.spec.priority)

        # Then connect it:
        if not no_root_connection:
            self.task_graph.add_edge(ROOT, task.name)

        self._insert_dependencies(task)
        self._insert_dependents(task)
        self._insert_according_to_queue_behaviour(task)

    def update_state(self, task:str|TaskBase_i|JGDVTaskArtifact, state:self.state_e):
        """ update the state of a task in the dependency graph """
        logging.debug("Updating Task State: %s -> %s", task, state)
        match task, state:
            case str(), self.state_e() if task in self.task_graph:
                self.task_graph.nodes[task]['state'] = state
            case TaskBase_i(), self.state_e() if task.name in self.task_graph:
                self.task_graph.nodes[task.name]['state'] = state
            case JGDVTaskArtifact(), self.state_e() if task in self.task_graph:
                self.task_graph.nodes[task]['state'] = state
            case _, _:
                raise doot.errors.DootTaskTrackingError("Bad task update state args", task, state)

    def next_for(self, target:None|str=None) -> None|Job_i|Task_i:
        """ ask for the next task that can be performed """
        if target and target not in self.active_set:
            self.queue_task(target, silent=True)

        focus : str | JGDVTaskArtifact | None = None
        adj                                   = dict(self.task_graph.adjacency())
        while bool(self.task_queue):
            focus : str = self.task_queue.peek()
            logging.debug("Task: %s  State: %s, Stack: %s", focus, self.task_state(focus), self.active_set)

            if focus in self.task_graph and self.task_graph.nodes[focus][PRIORITY] < self._min_priority:
                logging.warning("Task reached minimum priority while waiting, and has been cancelled: %s", focus)
                self.update_state(focus, self.state_e.FAILED)

            match self.task_state(focus):
                case self.state_e.SUCCESS: # remove task on completion
                    self.deque_task()
                case self.state_e.EXISTS:  # remove artifact when it exists
                    for pred in self.task_graph.pred[focus].keys():
                        logging.debug("Propagating Artifact existence to disable: %s", pred)
                        self.update_state(pred, self.state_e.SUCCESS)
                    self.deque_task()
                    return self.artifacts[focus]
                case self.state_e.HALTED:  # remove and propagate halted status
                    # anything that depends on a halted task in turn gets halted
                    halting = list(self.task_graph.succ[focus].keys())
                    printer.warning("Propagating Halt from: %s to: %s", focus, halting)
                    for pred in halting:
                        self.update_state(pred, self.state_e.HALTED)
                    # And remove the halted task from the active_set
                    self.deque_task()
                case self.state_e.FAILED:  # stop when a task fails, and clear any queued tasks
                    self.clear_queue()
                    return None
                case self.state_e.RUNNING:
                    logging.debug("Got Running Task: %s, continuing", focus)
                case self.state_e.READY if focus in self.execution_path: # error on running the same task twice
                    raise doot.errors.DootTaskTrackingError("Task Attempted to run twice: %s", focus)
                case self.state_e.READY:   # return the task if its ready
                    self.execution_path.append(focus)
                    self.update_state(focus, self.state_e.RUNNING)
                    self.task_graph.nodes[focus][PRIORITY] -= 1
                    return self.tasks.get(focus, None)
                case self.state_e.ARTIFACT if bool(self.artifacts[focus]): # if an artifact exists, mark it so
                    self.update_state(focus, self.state_e.EXISTS)
                case self.state_e.ARTIFACT: # Add dependencies of an artifact to the stack
                    incomplete, all_deps = self._task_dependencies(focus)
                    if bool(incomplete):
                        logging.info("Artifact Blocking Check: %s", focus)
                        self.deque_task()
                        self.queue_task(focus, *incomplete, silent=True)
                    elif bool(all_deps):
                        logging.debug("Artifact Unblocked: %s", focus)
                        self.update_state(focus, self.state_e.EXISTS)
                    else:
                        self.deque_task()
                        self.queue_task(focus)
                case self.state_e.WAIT | self.state_e.DEFINED: # Add dependencies of a task to the stack
                    incomplete, _ = self._task_dependencies(focus)
                    if bool(incomplete):
                        logging.info("Task Blocked: %s on : %s", focus, incomplete)
                        self.update_state(focus, self.state_e.WAIT)
                        self.deque_task()
                        self.queue_task(focus, *incomplete, silent=True)
                    else:
                        logging.debug("Task Unblocked: %s", focus)
                        self.update_state(focus, self.state_e.READY)

                case self.state_e.DECLARED: # warn on undefined tasks
                    logging.warning("Tried to Schedule a Declared but Undefined Task: %s", focus)
                    self.deque_task()
                    self.update_state(focus, self.state_e.SUCCESS)
                case _: # Error otherwise
                    raise doot.errors.DootTaskTrackingError("Unknown task state: ", x)

        return None



#!/usr/bin/env python3
"""


See EOF for license/metadata/notes as applicable
"""

##-- builtin imports
from __future__ import annotations

# import abc
import datetime
import enum
import functools as ftz
import itertools as itz
import logging as logmod
import pathlib as pl
import re
import time
import types
import weakref
# from copy import deepcopy
# from dataclasses import InitVar, dataclass, field
from typing import (TYPE_CHECKING, Any, Callable, ClassVar, Final, Generic,
                    Iterable, Iterator, Mapping, Match, MutableMapping,
                    Protocol, Sequence, Tuple, TypeAlias, TypeGuard, TypeVar,
                    cast, final, overload, runtime_checkable, Generator)
from uuid import UUID, uuid1

##-- end builtin imports

##-- lib imports
import more_itertools as mitz
##-- end lib imports

##-- logging
logging = logmod.getLogger(__name__)
##-- end logging

from jgdv._interfaces.decorator import JGDVDecorator_i
from jgdv.debugginer.running_debugger import RunningDebugger

class JGDVBreakpoint(JGDVDecorator_i):
    """
      Decorator to attach a debugger to a function, without pausing execution
    """

    def __call__(self, *args, **kwargs):
        # TODO handle repeats
        if args[0].breakpoint:
            f_code = f.__code__
            db = RunningDebugger()
            # Ensure trace function is set
            sys.settrace(db.trace_dispatch)
            if not db.get_break(f_code.co_filename, f_code.co_firstlineno+2):
                db.set_break(f_code.co_filename,
                            f_code.co_firstlineno+2,
                            True)
            else:
                bp = Breakpoint.bplist[f_code.co_filename,
                                    f_code.co_firstlineno+2][0]
                bp.enable()


        return self._func(self, *args, **kwargs)



#!/usr/bin/env python2
##-- imports
from __future__ import annotations

import pathlib as pl
from dataclasses import InitVar, dataclass, field
from typing import (Any, Callable, ClassVar, Generic, Iterable, Iterator,
                    Mapping, Match, MutableMapping, Sequence, Tuple, TypeAlias,
                    TypeVar, cast)

##-- end imports

from datetime import datetime, timedelta
from collections import defaultdict

class TimelineFile:
    """
    # Timeline Format:
    # Ext: .timeline

    # For file global tags:
    :tags ....

    # Form One: Event
    year       event country person* :desc .... :tags ... :wiki ....

    # Form Two: Period:
    year -> year event country person_surname* :tags ... :wiki .... :desc ....

    1922 -> 1933 "event"      england blah_bloo bloo_blee :tags blah,blee,blah :link blah
    2003         "something"  usa     a_person            :tags politics       :link https://blah :desc
    """

    date_dict : dict[datetime, list[TimelineEntry]]


    entries : dict[str, TimelineEntry]

    def __init__(self):
        self.date_dict = defaultdict(list)
        self.entries   = dict()
        self.min = None
        self.max = None

    @staticmethod
    def read(fpath:pl.Path):
        result = TimelineFile()
        for line in fpath.read_text().split("\n"):
            # parse line
            data = TimelineEntry()
            results.add(data)

    def __iter__(self):
        for key in sorted(self.date_dict.keys()):
            yield self.date_dict[key]

    def __str__(self):
        results = []
        for entry_group in self:
            results += map(str, entry_group)

        return "\n".join(results)

    def __repr__(self):
        return f"<{self.__class__.__name__}: {len(self)}>"

    def __contains__(self, value):
        match value:
            case datetime():
                return value in self.date_dict
            case TimelineEntry():
                return value.when in self.date_dict

    def __iadd__(self, values):
        match values:
            case [] | None:
                pass
            case [*args]:
                for arg in args:
                    assert(isinstance(arg, TimelineEntry))
                    self.add(arg)
            case TimelineEntry():
                self.add(values)
            case TimelineFile():
                for entrygroup in values:
                    self.add(*entrygroup)

        return self

    def __len__(self):
        return len(self.date_dict)

    def range(self):
        return (self.min, self.max)

    def encompasses(self, value):
        match value:
            case datetime():
                return self.min <= value <= self.max
            case TimelineEntry():
                valMin, valMax = value.range()
                return self.min <= valMin and valMax <= self.max
            case TimelineFile():
                return self.min <= value.min and value.max <= self.max

    def intersects(self, value):
        match value:
            case datetime():
                return self.encompasses(value)
            case TimelineEntry():
                valMin, valMax = value.range()
                return self.min <= valMin or valMax <= self.max
            case TimelineFile():
                left_intersect = self.min <= value.min and self.max <= value.max
                right_intersect = value.min <= self.min and value.max <= self.max
                return left_intersect or right_intersect



    def add(self, *data):
        for entry in data:
            assert(isinstance(entry, TimelineEntry))
            self.date_dict[entry.when].append(entry)
            match self.min, self.max:
                case None, None:
                    self.min = entry.when
                    self.max = entry.when
                case minDate, _ if entry.when < minDate:
                    self.min = entry.when
                case _, maxDate if maxDate < entry.when:
                    self.max = entry.when



"""
Utility decorators
"""
##-- imports
from __future__ import annotations

import logging as logmod
from functools import wraps
from typing import (Any, Callable, ClassVar, Dict, Generic, Iterable, Iterator,
                    List, Mapping, Match, MutableMapping, Optional, Sequence,
                    Set, Tuple, TypeVar, Union, cast)

logging = logmod.getLogger(__name__)

from typing import TYPE_CHECKING
if TYPE_CHECKING:
    # tc only imports
    pass

##-- end imports

from jgdv._interfaces.decorator import JGDVDecorator_i

class LogReturn(JGDVDelayDecorator_i):
    """
    Utility Decorator to log a functions return value at a set level
    """

    def __init__(self, prefix, level=logmod.DEBUG, msg=None, logger=None):
        super().__init__()
        self._prefix = prefix
        self._level  = level
        self._msg    =  msg or "{prefix} result: {result}"
        self._logger = logger or logmod.getLogger("jgdv._returns")

    def _wrapper(self, *args, **kwargs):
        result = self._func(*args, **kwargs)
        log_msg = self._msg.format(prefix=self._prefix, result=result)
        self._logger.log(level, log_msg)
        return result



"""
Provides default comparison functions to be used in the RBTree
Including standard <, > and ==,
along with arc comparisons for beachline use
"""
##-- imports
from __future__ import annotations

from enum import Enum
import logging as root_logger

##-- end imports

logging = root_logger.getLogger(__name__)
Directions = Enum('Directions', 'LEFT RIGHT')

#funcs take a node, a value, and data to contextualise

def default_comparison(a, b, comp_data):
    """ Standard smallest to largest comparison function """
    if a.value < b:
        return Directions.RIGHT
    return Directions.LEFT

def inverted_comparison(a, b, comp_data):
    """ Standard largest to smallest comparison function """
    if a.value < b:
        return Directions.LEFT
    return Directions.RIGHT

def default_equality(a, b, eq_data):
    """ Standard Equality test """
    return a.value == b


class ArcCmp:
    def arc_equality(a, b, eq_data):
        """ Compare two arcs for equality """
        #return true if  a.pred|a < b < a|a.succ
        l_inter, r_inter = __arc_intersects(a, b, eq_data)
        result = False
        if l_inter is not None and r_inter is not None:
            result = l_inter < b < r_inter
        elif r_inter is not None:
            result = b < r_inter
        elif l_inter is not None:
            result = l_inter < b
        return result

    def arc_comparison(a, b, comp_data):
        """ Function to compare an arc and xposition
        Used in Beachline/Voronoi """
        l_inter, r_inter = __arc_intersects(a, b, comp_data)
        pred_self = False
        self_succ = False

        if l_inter is None and r_inter is None: #Base case: single arc
            if b < a.value.fx:
                return Directions.LEFT
            else:
                return Directions.RIGHT

        if l_inter is not None:
            if b < l_inter:
                pred_self = True
        if r_inter is not None:
            if r_inter < b:
                self_succ = True

        if pred_self:
            return Directions.LEFT
        if self_succ:
            return Directions.RIGHT

        return Directions.LEFT

    def __arc_intersects(a, b, comp_data):
        """ Internal function to test if two arcs intersect """
        pred = a.getPredecessor()
        succ = a.getSuccessor()
        pred_intersect = None
        succ_intersect = None
        pred_intersect_out = None
        succ_intersect_out = None
        pred_above_self = None
        succ_above_self = None

        if pred is not None:
            pred_intersect = a.value.intersect(pred.value)
            pred_above_self = a.value.fy < pred.value.fy

        if succ is not None:
            succ_intersect = succ.value.intersect(a.value)
            succ_above_self = a.value.fy < succ.value.fy

        if pred_intersect is not None and len(pred_intersect) == 1:
            pred_intersect_out = pred_intersect[0, 0]
        elif pred_intersect is not None and len(pred_intersect) == 2:
            if pred_above_self:
                pred_intersect_out = pred_intersect[0, 0]
            else:
                pred_intersect_out = pred_intersect[1, 0]

        if succ_intersect is not None and len(succ_intersect) == 1:
            succ_intersect_out = succ_intersect[0, 0]
        elif succ_intersect is not None and len(succ_intersect) == 2:
            if succ_above_self:
                succ_intersect_out = succ_intersect[1, 0]
            else:
                succ_intersect_out = succ_intersect[0, 0]

        return (pred_intersect_out, succ_intersect_out)


class LineCmp:
    def line_cmp(a, b, cd):
        """ Line comparison to be used in the status tree """
        #Is horizontal:
        a_hor = a.value.is_flat()
        b_hor = b.is_flat()
        #---
        logging.debug("Comparison: {} - {}".format(a.value.index, b.index))
        logging.debug("Flat:       {} - {}".format(a_hor, b_hor))
        y = cd['y']
        if not (a_hor or b_hor):
            y += cd['nudge']
        a_ranges = a.value.get_ranges()
        b_ranges = b.get_ranges()

        a_val = a.value(y=y)[0]
        b_val = b(y=y)[0]
        if b_hor:
            b_val = min(max(cd['x'], b_ranges[0, 0]), b_ranges[0, 1])
        if a_hor:
            a_val = min(max(cd['x'], a_ranges[0, 0]), a_ranges[0, 1])

        logging.debug("Values: {} - {}".format(a_val, b_val))

        if a_val <= b_val:
            return Directions.RIGHT
        return Directions.LEFT

    def line_eq(a, b, cd):
        """ Compare two lines against each other """
        #pylint: disable=unused-argument
        return a.value == b

    def line_cmp_vert(a, b, cd):
        """ Compare a line against a point """
        result = Directions.LEFT
        if a.value.is_flat():
            a_val = cd['x']
        else:
            a_val = a.value(y=cd['y'])[0]
        logging.debug("VERT a_val{}: {}  b_val: {}".format(a.value.index, a_val, b[0]))

        if a_val <= b[0]:
            result = Directions.RIGHT
        return result

    def line_eq_vert(a, b, cd):
        """ test for equality between a line and a point """
        return np.allclose(a.value(y=cd['y']), b)



#!/usr/bin/env python3
"""

"""
##-- imports
from __future__ import annotations

import types
import abc
import datetime
import enum
import functools as ftz
import itertools as itz
import logging as logmod
import pathlib as pl
import re
import time
from copy import deepcopy
from dataclasses import InitVar, dataclass, field
from typing import (TYPE_CHECKING, Any, Callable, ClassVar, Final, Generic,
                    Iterable, Iterator, Mapping, Match, MutableMapping,
                    Protocol, Sequence, Tuple, TypeAlias, TypeGuard, TypeVar,
                    cast, final, overload, runtime_checkable)
from uuid import UUID, uuid1
from weakref import ref

##-- end imports

##-- logging
logging = logmod.getLogger(__name__)
##-- end logging

class VoronoiIO:

    @staticmmethod
    def save_graph(values:np.ndarray, fpath):
        """
        Pickle the voronoi points into a file
        """
        fpath.write_text(pickle.dumps(values)

    @staticmethod
    def load_graph(fpath):
        """
        Load the pickled graph of voronoi points
        """
        return pickle.loads(fpath.read_text())



#!/usr/bin/env python3
"""

"""
##-- imports
from __future__ import annotations

import abc
import logging as logmod
import pathlib as pl
from copy import deepcopy
from dataclasses import InitVar, dataclass, field
from re import Pattern
from typing import (TYPE_CHECKING, Any, Callable, ClassVar, Final, Generic,
                    Iterable, Iterator, Mapping, Match, MutableMapping,
                    Protocol, Sequence, Tuple, TypeAlias, TypeGuard, TypeVar,
                    cast, final, overload, runtime_checkable)
from uuid import UUID, uuid1
from weakref import ref

if TYPE_CHECKING:
    # tc only imports
    pass
##-- end imports

##-- logging
logging = logmod.getLogger(__name__)
# If CLI:
# logging = logmod.root
# logging.setLevel(logmod.NOTSET)
##-- end logging

def copy_files(source_dir:pl.Path, target_dir:pl.Path):
    logging.info("Copying from %s to %s", source_dir, target_dir)
    if source_dir.exists() and not target_dir.exists():
        logging.info("as group")
        result = run(['cp' ,'-r', str(source_dir), str(target_dir)], capture_output=True, check=True)
    elif target_dir.exists():
        logging.info("as individual")
        for y in source_dir.iterdir():
            if not y.is_file():
                continue

            call_sig = ['cp', str(y), str(target_dir / y.name)]
            run(call_sig, capture_output=True, check=True)

def copy_new(source:pl.Path, lib_path:pl.Path):
    logging.info("Adding to library with: %s", source)
    file_dir  = source.parent / f"{source.stem}_files"

    first_letter = source.name[0].lower()
    if not ("a" <= first_letter <= "z"):
        first_letter = "symbols"

    target_for_new = lib_path / f"group_{first_letter}"

    if not (target_for_new / source.name).exists():
        run(['cp', str(source), str(target_for_new)], capture_output=True, check=True)

    run(["mv", str(source), str(source.parent / f"{source.name}{PROCESSED}")], capture_output=True, check=True)

    copy_files(file_dir, target_for_new / f"{source.stem}_files")

def integrate(source:pl.Path, lib_dict:dict[str,Any]):
    logging.info("Integrating: %s", source)
    new_files      = source.parent / f"{source.stem}_files"
    existing_org   = lib_dict[source.name] /  source.name
    existing_files = lib_dict[source.name] / f"{source.stem}_files"

    assert(existing_org.exists())
    if not existing_files.exists():
        existing_files.mkdir()

    with open(source, 'r') as f:
        lines = f.read()

    with open(existing_org, 'a') as f:
        f.write("\n")
        f.write(lines)

    run(["mv", str(source), str(source.parent / f"{source.name}{PROCESSED}")], capture_output=True, check=True)

    if not new_files.exists():
        return

    copy_files(new_files, existing_files)

# def update_record(path:pl.Path, sources:List[pl.Path]):
#     now : str = datetime.datetime.now().strftime("%Y-%m-%d")

#     all_ids = get_all_tweet_ids(*sources, ext=".org_processed")

#     # add it to the library record
#     with open(path, 'a') as f:
#         # Insert date
#         f.write(f"{now}:\n\t")
#         f.write("\n\t".join(sorted(all_ids)))
#         f.write("\n----------------------------------------\n")



#/usr/bin/env python3
"""
How model results are represented and manipulated
"""
##-- imports
from __future__ import annotations

import abc
import logging as logmod
from collections.abc import Sequence
from copy import deepcopy
from dataclasses import InitVar, dataclass, field
from re import Pattern
from typing import (TYPE_CHECKING, Any, Callable, ClassVar, Final, Generic,
                    Iterable, Iterator, Mapping, Match, MutableMapping,
                    Protocol, Tuple, TypeAlias, TypeGuard, TypeVar,
                    cast, final, overload, runtime_checkable)
from uuid import UUID, uuid1
from weakref import ref

##-- end imports

logging = logmod.getLogger(__name__)

from clingo import Symbol
from jgdv.apis.clingo.ast import SolverAST
from jgdv.apis.clingo.solver import SolverModelResult

STATE_HOLDSAT_GROUPS = []

@dataclass
class State_i:
    """
    Description of a single moment in a model's trace.
    Depending on implementation, could hold SolverAST's
    or Clingo Symbols
    """

    timestep : int            = field(default=0)
    holdsat  : dict[str, Any] = field(default_factory=dict)
    occurred : list[Any]      = field(default_factory=list)
    observed : list[Any]      = field(default_factory=list)
    rest     : list[Any]      = field(default_factory=list)

    def __post_init__(self):
        for x in STATE_HOLDSAT_GROUPS:
            self.holdsat[x] = []

    def __iter__(self):
        for x in self.fluents:
            yield x

        for x in self.occurred:
            yield x

        for x in self.observed:
            yield x

        for x in self.rest:
            yield x

    @property
    def fluents(self) -> iter[Any]:
        """ Iterate through all the fluents as a list
        instead of as a dict of separate types of fluent
        """
        for items in self.holdsat.values():
            for entry in items:
                yield entry

    @abc.abstractmethod
    def __repr__(self): pass

    @abc.abstractmethod
    def to_json(self) -> dict: pass

    @abc.abstractmethod
    def check(self, conditions) -> bool: pass

    @abc.abstractmethod
    def insert(self, val:str|SolverAST|Symbol): pass

    @abc.abstractmethod
    def filter(self, allow:list[Any], reject:list[Any]) -> State_i: pass

    # @abc.abstractmethod
    # def conflicts(self) -> list[tuple[int, Any, Any]]: pass

@dataclass
class Trace_i(Sequence):
    """
    The collected sequence of instance states which comprise
    a full model run
    """
    states   : InitVar[list[State_i]] = field()
    metadata : dict                   = field(default_factory=dict)

    _states  : dict[str, State_i]     = field(default_factory=dict)
    _ordered : list[int]              = field(init=False, default_factory=list)
    state_constructor : ClassVar[State_i] = None

    def __post_init__(self, states):
        self._states  = {x.timestep : x for x in states}
        self._ordered = sorted(x.timestep for x in states)

    @staticmethod
    @abc.abstractmethod
    def from_json(data): pass

    @staticmethod
    @abc.abstractmethod
    def from_model(model:SolverModelResult, steps:int=1): pass

    def __getitem__(self, index):
        return self._states[index]

    def __iter__(self):
        return iter(self._states.values())

    def contextual_iter(self) -> iter[tuple]:
        """
        provide an iterator of tuples
        [timestep, state, state-1, state+1]
        """
        states : list = list(self._states.values())
        return zip(range(len(self)),
                   states,
                   [None] + states,
                   states[1:] + [None])

    def __len__(self):
        return len(self._states)

    def last(self) -> State_i:
        return self._states[str(self._ordered[-1])]

    @property
    def timesteps(self) -> list[int]:
        return list(self._states.keys())

    @abc.abstractmethod
    def __repr__(self): pass

    @abc.abstractmethod
    def check(self, conditions:list) -> bool: pass

    @abc.abstractmethod
    def to_json_str(self, filename=None) -> str: pass

    @abc.abstractmethod
    def filter(self, allow:list[str], reject:list[str], start:None|int=None, end:None|int=None) -> Trace_i: pass

    @abc.abstractmethod
    def fluent_intervals(self) -> list[tuple[str, int, int]]: pass

    # @abc.abstractmethod
    # def conflicts(self) -> list[tuple[int, Any, Any]]: pass

@dataclass
class TraceTree_i:
    """
    Collections of traces which merge to form a tree, branching at diverging actions
    """

    # [state, [successors]]
    states : dict[str, tuple[State_i, list[str]]]
    start  : str

    def add_trace(self, trace):
        """ Add a trace into the tree
        comparing and discarding states until the divergent point is found.
        """
        pass



"""

"""
##-- imports
from __future__ import annotations

import abc
import importlib
import logging as logmod
import re
from collections import defaultdict
from dataclasses import InitVar, dataclass, field
from datetime import datetime
from enum import Enum
from os.path import abspath, exists, expanduser, split, splitext
from typing import (TYPE_CHECKING, Any, Callable, ClassVar, Final, Generic,
                    Iterable, Iterator, Mapping, Match, MutableMapping,
                    Protocol, Sequence, Tuple, TypeAlias, TypeGuard, TypeVar,
                    cast, final, overload, runtime_checkable)

##-- end imports

import pyparsing as pp

config = acab.config

logging = logmod.getLogger(__name__)

# TODO shift this into config
ModuleFragment : TypeAlias = AT.ModuleFragment

SPLIT_RE         = re.compile("[ .!?/]")
shortcut_config  = config.module.REPL.shortcuts
shortcut_pairs   = sorted([(shortcut_config[cmd], cmd) for cmd in shortcut_config._keys])

@register
def do_shortcuts(self, line):
    """
    Print the :{kw} shortcut bindings loaded from config
    """
    print("Repl Shortcut commands: ")
    for kw, cmd in shortcut_pairs:
        print(f"    :{kw:<5} -> {cmd}")

@register
def do_acab(self, line):
    """ All Cops Are Bastards """
    print("All Cops Are Bastards")



#!/usr/bin/env python3
"""


See EOF for license/metadata/notes as applicable
"""

##-- builtin imports
from __future__ import annotations

# import abc
import datetime
import enum
import functools as ftz
import itertools as itz
import logging as logmod
import pathlib as pl
import re
import time
import types
import weakref
# from copy import deepcopy
# from dataclasses import InitVar, dataclass, field
from typing import (TYPE_CHECKING, Any, Callable, ClassVar, Final, Generic,
                    Iterable, Iterator, Mapping, Match, MutableMapping,
                    Protocol, Sequence, Tuple, TypeAlias, TypeGuard, TypeVar,
                    cast, final, overload, runtime_checkable, Generator)
from uuid import UUID, uuid1

##-- end builtin imports

##-- lib imports
import more_itertools as mitz
##-- end lib imports

##-- logging
logging = logmod.getLogger(__name__)
##-- end logging

from jgdv.files.org.base import OrgBuilderBase
from jgdv.files.org.drawer import OrgDrawerBuilder

@dataclass
class OrgStrBuilder(OrgBuilderBase):
    """
    Utility class for building Org files
    """
    output : List[Union[str, 'OrgBuilderBase']] = field(default_factory=list)

    prop_align_len      : int = 10
    heading_char        : str = "*"

    def heading(self, level, *text):
        stars = level * self.heading_char
        self.add(" ".join([stars, *text]))

    def link(self, text, uri):
        self.add(self.named_link_pattern.format(text, uri))
        self.nl

    def links(self, links):
        converted = [self.link_pattern.format(x) for x in links]
        self.add(*converted)

    def add(self, *text):
        self.output += text

    def drawer(self, name):
        drawer = OrgDrawerBuilder(self, name)
        self.add(drawer)
        return drawer

    @property
    def nl(self):
        self.output.append("")

    def __str__(self):
        return "\n".join(map(str, self.output))



"""
A Function to create a heightmap, quantized,
uses the red channel
"""
import numpy as np
from scipy.ndimage import generic_filter
from functools import partial
from noise import snoise2

import logging as root_logger
logging = root_logger.getLogger(__name__)

baseheight = np.array([0.3, 0, 0, 0])

class HeightMap:

    @staticmethod
    def build_with_edges(size, minheight, layers, oct=3, repeatx=None, repeaty=None, base=0.0):
        """ Generates a heightmap, and the edges """
        hm    = HeightMap.build(size, minheight, oct=oct, repeatx=repeatx, repeaty=repeaty, base=base)
        qhm   = quantize(hm, layers)
        edges = HeightMap.detect_edges(qhm, layers)
        return (hm, qhm, edges)

    @staticmethod
    def build(size, minheight, oct=3, repeatx=None, repeaty=None, base=0.0):
        """ Generate a height map """
        matrix = np.zeros((size,size))
        SCALER = 1 / size

        for x in range(0, size):
            for y in range(0, size):
                xs = x * SCALER
                ys = y * SCALER
                value = snoise2(xs, ys, octaves=oct, repeatx=repeatx,
                                repeaty=repeaty, base=base)
                matrix[x,y] = value

        mmin = matrix.min()
        if mmin < minheight:
            matrix += minheight - mmin
            matrix /= matrix.max()

        return matrix

    @staticmethod
    def detect_edges(matrix, layers):
        """ Given a quantized heightmap and the number of steps,
        find the edges using convolution """
        logging.info("Detecting Edges")
        edges = np.zeros(matrix.shape)
        scaling = 100
        scaling_recip = 1 / scaling
        quantize_base = int(scaling / layers)
        scaled_matrix = matrix * scaling

        #For each level of quantization:
        for index in range(layers, 1, -1):
            layer_val = index * quantize_base
            #merge with layers above and filter
            use_matrix = scaled_matrix * (scaled_matrix >= layer_val)
            filter_mat = generic_filter(use_matrix, partial(HeightMap.filter_func, layer_val), size=2)
            #aggregate:
            edges += filter_mat

        return edges

    @staticmethod
    def filter_func(max_val, arr):
        """ A partial filter function to detect edges """
        return 0 in arr and any(arr >= max_val)

def gen_heightmap_and_edges(size, minheight, layers, oct=3, repeatx=None,
                            repeaty=None, base=0.0):
    """ Generates a heightmap, and the edges """
    hm = heightmap(size, minheight, oct=oct, repeatx=repeatx,
                   repeaty=repeaty, base=base)
    qhm = quantize(hm, layers)
    edges = detect_edges(qhm, layers)
    return (hm, qhm, edges)

def heightmap(size, minheight, oct=3, repeatx=None, repeaty=None, base=0.0):
    """ Generate a height map """
    matrix = np.zeros((size,size))
    SCALER = 1 / size

    for x in range(0, size):
        for y in range(0, size):
            xs = x * SCALER
            ys = y * SCALER
            value = snoise2(xs, ys, octaves=oct, repeatx=repeatx,
                            repeaty=repeaty, base=base)
            matrix[x,y] = value

    mmin = matrix.min()
    if mmin < minheight:
        matrix += minheight - mmin
        matrix /= matrix.max()

    return matrix

def quantize(matrix, layers):
    """ Given a basic heightmap, quantize it """
    logging.info("Quantizing")
    #pad the size a bit
    qmatrix = np.zeros(matrix.shape)
    scaling = 100
    scaling_recip = 1 / scaling
    quantize_base = int(scaling / layers)
    scaled_matrix = matrix * scaling

    for l in range(1,layers):
        quantize_value = quantize_base * l
        qmatrix += (scaled_matrix > quantize_value)

    qmatrix *= quantize_base
    qmatrix *= scaling_recip
    return qmatrix

def detect_edges(matrix, layers):
    """ Given a quantized heightmap and the number of steps,
    find the edges using convolution """
    logging.info("Detecting Edges")
    edges = np.zeros(matrix.shape)
    scaling = 100
    scaling_recip = 1 / scaling
    quantize_base = int(scaling / layers)
    scaled_matrix = matrix * scaling

    #For each level of quantization:
    for index in range(layers, 1, -1):
        layer_val = index * quantize_base
        #merge with layers above and filter
        use_matrix = scaled_matrix * (scaled_matrix >= layer_val)
        filter_mat = generic_filter(use_matrix, partial(filter_func, layer_val), size=2)
        #aggregate:
        edges += filter_mat

    return edges

def filter_func(max_val, arr):
    """ A partial filter function to detect edges """
    return 0 in arr and any(arr >= max_val)



##-- imports
from __future__ import annotations

# import abc
# import datetime
import enum
import functools as ftz
import itertools as itz
import logging as logmod
import pathlib as pl
import re
import time
import types
# from copy import deepcopy
# from dataclasses import InitVar, dataclass, field
from typing import (TYPE_CHECKING, Any, Callable, ClassVar, Final, Generic,
                    Iterable, Iterator, Mapping, Match, MutableMapping,
                    Protocol, Sequence, Tuple, TypeAlias, TypeGuard, TypeVar,
                    cast, final, overload, runtime_checkable)
# from uuid import UUID, uuid1
# from weakref import ref

##-- end imports

##-- logging
logging = logmod.getLogger(__name__)
##-- end logging

from tomlguard import TomlGuard
from collections import ChainMap
import jgdv

SEP : Final[str]          = "--"
PARAM_ASSIGN_PREFIX       = doot.constants.patterns.PARAM_ASSIGN_PREFIX
NON_DEFAULT_KEY           = doot.constants.misc.NON_DEFAULT_KEY

default_task : Final[str] = doot.config.on_fail((None,)).general.settings.default_task()
default_cmd  : Final[str] = doot.config.on_fail("run", str).general.settings.default_cmd()

class ArgParser_i:
    """
    A Single standard process point for turning the list of passed in args,
    into a dict, into a tomlguard,
    along the way it determines the cmds and tasks that have been chosne
    """

    def __init__(self):
        self.specs = []

    def add_param_specs(self, specs:list):
        self.specs += specs

    @abstractmethod
    def parse(self, args:list[str], doot_arg_specs:list[DootParamSpec], cmds:TomlGuard, tasks:TomlGuard) -> TomlGuard:
        raise NotImplementedError()

@jgdv.check_protocol
class JGDVCLIParser(ArgParser_i):
    """
    convert argv to tomlguard by:
    parsing each arg as toml,

    # doot {args} {cmd} {cmd_args}
    # doot {args} [{task} {tasks_args}] - implicit do cmd
    """

    class _ParseState(enum.Enum):
        HEAD  = enum.auto()
        CMD   = enum.auto()
        TASK  = enum.auto()
        EXTRA = enum.auto()

    def __init__(self):
        self.PS               = JGDVCLIParser._ParseState
        self.head_arg_specs   = None
        self.registered_cmds  = None
        self.registered_tasks = None
        self.default_help     = DootParamSpec(name="help", default=False, prefix="--")

        ## -- results
        self.head_call                          = None
        self.head_args                          = None
        self.cmd_name                           = None
        self.cmd_args                           = {}
        self.non_default_cmd_args               = []
        self.tasks_args                         = []
        self.extras                             = {}
        self.force_help                         = False

        ## -- loop state
        self.focus                             = self.PS.HEAD

    def _build_defaults_dict(self, param_specs:list) -> dict:
        return { x.name : x.default for x in param_specs }

    def parse(self, args:list, *, doot_specs:list[DootParamSpec], cmds:TomlGuard, tasks:TomlGuard) -> None|TomlGuard:
        """
          Parses the list of arguments against available registered parameter specs, cmds, and tasks.
        """
        logging.debug("Parsing args: %s", args)
        self.head_call                  = args[0]
        self.head_args                  = self._build_defaults_dict(doot_specs)
        self.head_arg_specs             = doot_specs
        self.registered_cmds            = cmds
        self.registered_tasks           = tasks
        self.focus                      = self.PS.HEAD

        remaining                      = args[1:]
        while bool(remaining):
            match self.focus:
                case self.PS.HEAD:
                    remaining = self.process_head(remaining)
                    self.focus = self.PS.CMD
                case self.PS.CMD:
                    remaining = self.process_cmd(remaining)
                    self.focus = self.PS.TASK
                case self.PS.TASK:
                    remaining = self.process_task(remaining)
                    self.focus = self.PS.EXTRA
                case self.PS.EXTRA:
                    remaining = self.process_extra(remaining)


        if self.cmd_args.get('help', False) is True:
            self.cmd_args['target']      = self.cmd_name
            self.cmd_name                = "help"
        elif any(x[1].get('help', False) is True for x in self.tasks_args if (target:=x[0])):
            self.cmd_args['target'] = target
            self.cmd_name = "help"

        # TODO ensure duplicated tasks have different args
        data = {
            "head"   : {"name": self.head_call, "args": self.head_args },
            "cmd"    : {"name" : self.cmd_name, "args" : self.cmd_args, NON_DEFAULT_KEY: self.non_default_cmd_args },
            "tasks"  : { name : args for name,args in self.tasks_args  },
            "extras" : self.extras
            }
        return TomlGuard(data)

    def process_head(self, args) -> list[str]:
        """ consume arguments for doot actual """
        logging.debug("Head Parsing: %s", args)
        head = args[0]
        while bool(args) and args[0] not in self.registered_cmds and args[0] not in self.registered_tasks:
            match [x for x in self.head_arg_specs if x == args[0]]:
                case []:
                    raise doot.errors.DootParseError("Unrecognized head arg", args[0])
                case [x]:
                    x.maybe_consume(args, self.head_args)
                case [*xs]:
                    raise doot.errors.DootParseError("Multiple possible head args", args[0])

        return args

    def process_cmd(self, args) -> list[str]:
        """ consume arguments for the command being run """
        logging.debug("Cmd Parsing: %s", args)
        head                     = args[0]
        cmd                      = self.registered_cmds.get(head, None)
        self.cmd_name            = head
        if cmd is None:
            cmd                      = self.registered_cmds[default_cmd]
            self.cmd_name            = default_cmd
            current_specs            = list(sorted(cmd.param_specs, key=DootParamSpec.key_func))
            self.cmd_args            = self._build_defaults_dict(current_specs)
            return args

        current_specs            = list(sorted(cmd.param_specs, key=DootParamSpec.key_func))
        self.cmd_args            = self._build_defaults_dict(current_specs)

        args.pop(0)
        while bool(args) and args[0] not in self.registered_tasks:
            if args[0] == SEP: # hit SEP, the forced separator
                # eg: doot list a b c -- something else
                args.pop(0)
                break
            match [x for x in current_specs if x == args[0]]:
                case []:
                    raise doot.errors.DootParseError("Unrecognized cmd arg", head, args[0])
                case [x]:
                    x.maybe_consume(args, self.cmd_args)
                case [*xs] if all(y.positional for y in xs):
                    self._consume_next_positional(args, self.cmd_args, xs)
                case [*xs] if len(y for y in xs if not y.positional):
                    raise doot.errors.DootParseError("Multiple possible cmd args", head, args[0])

        self.non_default_cmd_args = self._calc_non_default(self._build_defaults_dict(current_specs), self.cmd_args)
        return args

    def process_task(self, args) -> list[str]:
        """ consume arguments for tasks """
        logging.debug("Task Parsing: %s", args)
        if args[0] not in self.registered_tasks:
            task                     = self.registered_tasks[default_task]
            assert(isinstance(task, DootTaskSpec))
            task_name                 = default_task
            spec_params               = [DootParamSpec.from_dict(x) for x in task.extra.on_fail([], list).cli()]
            ctor_params               = task.ctor.try_import().param_specs
            current_specs             = list(sorted(spec_params + ctor_params, key=DootParamSpec.key_func))
            task_args                 = self._build_defaults_dict(current_specs)
            task_args[NON_DEFAULT_KEY] = []
            self.tasks_args.append((task_name, task_args))
            return args


        while bool(args) and args[0] in self.registered_tasks:
            task_name                 = args.pop(0)
            task                      = self.registered_tasks[task_name]
            assert(isinstance(task, DootTaskSpec))
            spec_params               = [DootParamSpec.from_dict(x) for x in task.extra.on_fail([], list).cli()]
            ctor_params               = task.ctor.try_import().param_specs
            current_specs             = list(sorted(spec_params + ctor_params, key=DootParamSpec.key_func))
            task_args                 = self._build_defaults_dict(current_specs)
            default_args              = task_args.copy()
            logging.debug("Parsing Task args for: %s: Available: %s", task_name, task_args.keys())

            while bool(args) and args[0] not in self.registered_tasks:
                if args[0] == SEP:
                    args.pop(0)
                    break

                match [x for x in current_specs if x == args[0]]:
                    case [] if args[0].startswith(PARAM_ASSIGN_PREFIX):
                        # No matches, its a free cli arg.
                        try:
                            key, *values     = args[0].split("=")
                            task_args[key.removeprefix("--")] = values
                            args.pop(0)
                        except ValueError:
                            raise doot.errors.DootParseError("Arg failed to split into key=value", args[0])
                    case [x] if not x.positional:
                        x.maybe_consume(args, task_args)
                    case [*xs] if all(y.positional for y in xs):
                        self._consume_next_positional(args, task_args, xs)
                    case [*xs]:
                        raise doot.errors.DootParseError("Multiple possible task args", task_name, args[0])


            if task_name in [x[0] for x in self.tasks_args]:
                raise doot.errors.DootParseError("a single task was specified twice")

            task_args[NON_DEFAULT_KEY] = self._calc_non_default(default_args, task_args)
            logging.debug("Resulting Arg Assignments for %s : %s", task_name, task_args)
            self.tasks_args.append((task_name, task_args))

        return args

    def process_extra(self, args) -> None:
        logging.debug("Extra Parsing: %s", args)
        return None


    def _consume_next_positional(self, args, arg_dict, params:list):
        """
          try each positional param until success
        """
        for param in params:
            try:
                if 1 <= param.maybe_consume(args, arg_dict):
                    return
            except doot.errors.DootParseResetError:
                continue

        raise doot.errors.DootParseError("No positional argument succeeded", args)

    def _calc_non_default(self, defaults, actual) -> list:
        return [x for x,y in actual.items() if x not in defaults or defaults[x] != y]



"""
Utility Module to simplify making a GIF from individual images
"""
#pylint: disable=too-many-arguments
import logging as root_logger
from os.path import isfile, join
from os import listdir
import re
import imageio

logging = root_logger.getLogger(__name__)

class MakeGif:
    """ A Utility class to easily create gifs from a number of images """

    def __init__(self, output_dir=".",
                 gif_name="anim.gif",
                 source_dir="images",
                 file_format=".png", fps=12):
        self.output_dir = output_dir
        self.gif_name = gif_name
        self.source_dir = source_dir
        self.file_format = file_format
        self.fps = fps

        self.num_regex = re.compile(r'(\d+)')

    def get_num(self, s):
        """ Given a String, extract a number from it,
        or return a default """
        logging.info("Getting num of: {}".format(s))
        assert(isinstance(s, str))
        #pylint: disable=broad-except
        try:
            return int(self.num_regex.search(s).group(0))
        except Exception:
            return 9999999

    def run(self):
        """ Trigger the creation of the GIF """
        # Get all Files
        files = [x for x in listdir(self.source_dir) if isfile(join(self.source_dir, x))]
        assert(bool(files))
        logging.info("Making gif of {} frames".format(len(files)))

        # Sort by the number extracted from the filename
        files.sort(key=self.get_num)

        # Export as a Gif
        logging.info("Starting GIF writing")
        with imageio.get_writer(join(self.output_dir, self.gif_name), mode='I') as writer:
            for filename in files:
                image = imageio.imread(join(self.source_dir, filename))
                writer.append_data(image)
        logging.info("Finished GIF writing")



#!/usr/bin/env python3
"""

See EOF for license/metadata/notes as applicable
"""

##-- builtin imports
from __future__ import annotations

import abc
import datetime
import enum
import functools as ftz
import itertools as itz
import logging as logmod
import pathlib as pl
import re
import time
import types
import weakref
# from copy import deepcopy
from dataclasses import InitVar, dataclass, field
from typing import (TYPE_CHECKING, Any, Callable, ClassVar, Final, Generic,
                    Iterable, Iterator, Mapping, Match, MutableMapping,
                    Protocol, Sequence, Tuple, TypeAlias, TypeGuard, TypeVar,
                    cast, final, overload, runtime_checkable, Generator)
from uuid import UUID, uuid1

##-- end builtin imports

##-- lib imports
import more_itertools as mitz
##-- end lib imports

##-- logging
logging = logmod.getLogger(__name__)
##-- end logging

import importlib
from tomlguard import TomlGuard
from jgdv.location.toml_loc import TomlLocation

PAD           : Final[int]   = 15
ARTIFACT      : Final[str]   = "!!Artifact!!"

@dataclass
class JGDVArtifact:
    """
      Wraps a toml defined location into an artifact

      Describes an artifact a task can produce or consume.
    Artifacts can be Definite (concrete path) or indefinite (glob path)
    """
    base     : TomlLocation = field()
    key      : DootKey      = field()

    @staticmethod
    def build(data:str|dict|pl.Path) -> JGDVArtifact:
        match data:
            case str() if data.startswith(doot.constants.patterns.FILE_DEP_PREFIX):
                base = TomlLocation.build(ARTIFACT, data.removeprefix(doot.constants.patterns.FILE_DEP_PREFIX))
                base.meta |= LocationMeta.file
                key = DootKey.build(base.base)
                return JGDVArtifact(base, key)
            case str():
                base = TomlLocation.build(ARTIFACT, data)
                key  = DootKey.build(base.base)
                return JGDVArtifact(base, key)
            case dict():
                base = TomlLocation.build(ARTIFACT, data)
                if "*" in str(base.base):
                    base.meta |= LocationMeta.indefinite
                key = DootKey.build(base.base)
                return JGDVArtifact(base, key)
            case pl.Path():
                base = TomlLocation.build(ARTIFACT, data)
                if "*" in str(base.base):
                    base.meta |= LocationMeta.indefinite
                key = DootKey.build(base.base)
                return JGDVArtifact(base, key)
            case _:
                raise TypeError("Unknown Type to build Artifact from: %s", data)

    def __repr__(self):
        return f"<TaskArtifact: {self.key} : {self.base.meta}>"

    def __str__(self):
        return str(self.base.base)

    def __hash__(self):
        return hash(str(self))

    def __eq__(self, other:JGDVArtifact|Any):
        match other:
            case JGDVArtifact():
                return self.base == other.base
            case _:
                return False

    def __bool__(self):
        return self.exists

    def __contains__(self, other):
        """ whether a definite artifact is matched by self, an indefinite artifact
          a/b/c.py ∈ a/b/*.py
          ________ ∈ a/*/c.py
          ________ ∈ a/b/c.*
          ________ ∈ a/*/c.*
          ________ ∈ **/c.py

        """
        if not self.check(LocationMeta.indefinite):
            return False

        for x,y in zip(self.parent.parts, other.parent.parts):
            if x == "**" or y == "**":
                break
            if x == "*" or y == "*":
                continue
            if x != y:
                return False

        suffix      = (not self._definite_suffix) or self.base.base.suffix == other.base.base.suffix
        stem        = (not self._definite_stem)    or self.base.base.stem == other.base.base.stem
        return  suffix and stem

    @property
    def exists(self):
        if not self.is_definite:
            return False
        as_path = self.key.to_path(None, None)
        return as_path.exists()

    @property
    def is_definite(self):
        """ tests the entire artifact path """
        return not self.check(LocationMeta.indefinite)

    @property
    def _definite_stem(self):
        """ tests the stem of the artifact """
        return "*" not in self.base.base.stem

    @property
    def _definite_suffix(self):
        """ tests the suffix of the artifact """
        return "*" not in self.base.base.suffix

    @property
    def parent(self):
        return self.base.base.parent

    def is_stale(self) -> bool:
        """ whether the artifact itself is stale """
        raise NotImplementedError('TODO')

    def check(self, meta:LocationMeta) -> bool:
        return self.base.check(meta)



""" Line: Representation of a 2d line """
##-- imports
from __future__ import annotations
import logging as root_logger
from dataclasses import InitVar, dataclass, field
from math import sqrt
from typing import (Any, Callable, ClassVar, Dict, Generic, Iterable, Iterator,
                    List, Mapping, Match, MutableMapping, Optional, Sequence,
                    Set, Tuple, TypeVar, Union, cast)

import numpy as np

##-- end imports

from typing import Final

logging = root_logger.getLogger(__name__)

CENTRE : Final = np.array([[0.5, 0.5]])

@dataclass
class Line:
    """
    A line as a start position, a unit direction, and a length,
    useful for algebraic manipulation
    """

    source               : np.ndarray = field()
    direction            : np.ndarray = field()
    length               : float      = field()
    #slope and intersect :
    slope                : float      = field()
    slope_intersect      : np.ndarray = field()
    swapped              : bool       = field(default=False)
    origin_array         : np.ndarray = field(default=None)

    @staticmethod
    def build(a):
        """
        Create a new line from two vertices
        """
        assert(isinstance(a, np.ndarray))
        assert(a.shape == (2, 2))
        #Calculate the line parameters:
        vec = a[1] - a[0]
        l = sqrt(pow(vec, 2).sum())
        scale = 0
        if l != 0:
            scale = 1/l
        d = vec * scale
        #cx = a.x + (dx * l)
        #Slope and intersect:
        q = a[1] - a[0]
        if q[0] == 0:
            m = None
            b = None
        else:
            m = q[1] / q[0]
            b = a[0, 1] - (m * a[0, 0])
        return Line(a[0], d, l, m, b, origin_array=a)

    def __post_init__(self):
        assert(all([isinstance(x, np.ndarray) for x in [s, d]]))

    def __repr__(self):
        return "<Line S: {}, D: {}, L: {}, SW: {}>".format(self.source,
                                                           self.direction,
                                                           self.length,
                                                           self.swapped)

    def __call__(self, x=None, y=None):
        """
        Solve the line for the None value.
        must have either x or y passed in
        """
        assert(any([a is not None for a in [x, y]]))
        assert(not all([a is not None for a in [x, y]]))
        if x is not None:
            if self.slope is not None:
                yprime = self.slope * x + self.slope_intersect
            else:
                yprime = self.source[1]
            return np.array([x, yprime])
        else:
            if self.slope is not None and self.slope != 0:
                xprime = (y - self.slope_intersect) / self.slope
            else:
                xprime = self.source[0]
            return np.array([xprime, y])

    def constrain(self, min_x, min_y, max_x, max_y):
        """
        Intersect the line with a bounding box, adjusting points as necessary
        """
        #min and max: [x, y]
        dest = self.destination()
        nline = np.row_stack((self.source, dest))

        xs_min = nline[:, 0] < min_x
        xs_max = nline[:, 0] > max_x
        ys_min = nline[:, 1] < min_y
        ys_max = nline[:, 1] > max_y
        xs = (nline[:, 0] * np.invert(xs_min + xs_max)) + (min_x * xs_min) + (max_x * xs_max)
        ys = (nline[:, 1] * np.invert(ys_min + ys_max)) + (min_y * ys_min) + (max_y * ys_max)
        return np.column_stack((xs, ys))

    def subdivide(self, n):
        """
        Subdivide this line into n subdivisions
        """
        assert(isinstance(n, int))
        #plus two
        subdivisions = np.linspace(0, 1, n+2).reshape((-1, 1))
        new_points = self.source + ((subdivisions * self.length) * self.direction)
        return new_points

    def ratio_subdivide(self, ss, srange=None):
        """
        subdivides a line by a set of ratios that sum to 1
        """
        assert(isinstance(ss, np.ndarray))
        assert(0.99 <= ss.sum() < 1.1)
        if srange is None:
            srange = (0, 1)
        assert(isinstance(srange, tuple))
        assert(srange[0] <= srange[1])

        new_points = np.array([self.source])
        current_r = 0.0
        ratios = list(ss[0])
        while bool(ratios):
            r = ratios.pop(0)
            if srange[1] < r:
                ratios.append(r - srange[1])
                r = srange[1]
            current_r += r
            if bool(ratios) and r <= srange[0]:
                continue
            if not bool(ratios) and current_r != 1:
                current_r = 1
            new_points = np.row_stack((new_points,
                                       self.source + ((current_r * self.length) * self.direction)))

        return new_points

    def destination(self, l=None, r=None):
        """
        Calculate the destination vector of the line
        """
        if r is not None:
            l = self.length * r
        if l is None:
            l = self.length
        return self.source + (l * self.direction)

    def bounds(self):
        """
        Get the start and end points of this line
        """
        if self.swapped:
            return np.row_stack((self.destination(), self.source))
        else:
            return np.row_stack((self.source, self.destination()))

    def intersect_with_circle(self, centre, radius):
        """
        Get the point that this line crosses a circle
        from http://csharphelper.com/blog/2014/09/determine-where-a-line-intersects-a-circle-in-c/
        """
        #pylint: disable=invalid-name
        A = sum(pow(self.direction, 2))
        B = 2 * sum(self.direction * (self.source - centre))
        C = sum(pow(self.source - centre, 2)) - pow(radius, 2)

        det = pow(B, 2) - (4 * A * C)

        if np.isclose(A, 0):
            raise Exception("No Intersection")

        if np.isclose(det, 0):
            t = -B / (2 * A)
            result = np.array([self.source + (t * self.direction)])
            return result

        #two intersections:
        t = (-B + sqrt(det)) / (2 * A)
        t2 = (-B - sqrt(det)) / (2 * A)

        result = self.source + (np.array([[t], [t2]]) * self.direction)
        return result

    def intersect(self, other):
        """ Intersect a line with another """
        #TODO implement line intersection
        return False



#!/usr/bin/env python3
##-- imports
from __future__ import annotations

import abc
import logging as logmod
from dataclasses import InitVar, dataclass, field
from typing import (TYPE_CHECKING, Any, Callable, ClassVar, Final, Generic,
                    Iterable, Iterator, Mapping, Match, MutableMapping,
                    Protocol, Sequence, Tuple, TypeAlias, TypeGuard, TypeVar,
                    cast, final, overload, runtime_checkable)

##-- end imports

logging = logmod.getLogger(__name__)

@register_class("tutorial")
class TutorialCmd:
    """
    Print out a basic tutorial of Acab and this Repl
    """

    def __init__(self):
        self._parser = self._gen_parser()

    def _gen_parser(self):
        pass

    def __call__(self, line):
        # Print a section, return to main loop,
        # if tutorial is called again, continue
        # if restart is passed in, restart the tutorial
        # also include option to print tutorial for a module
        return



#!/usr/bin/env python3
"""
Provides Metaclasses for creating Singletons

Note: superclass is type(Protocol) so classes which
implement protocol's don't get a metaclass conflict
"""
##-- imports
import logging as logmod
from fractions import Fraction
from typing import (Any, Callable, ClassVar, Dict, Generic, Iterable, Iterator,
                    List, Mapping, Match, MutableMapping, Optional, Sequence, Protocol,
                    Set, Tuple, TypeAlias, TypeVar, Union, cast)

##-- end imports

logging = logmod.getLogger(__name__)

class SingletonMeta(type(Protocol)):
    """
    Create an instance field to hold the singleton
    For Each Class Hierarchy
    """

    def __init__(cls, name:str, bases:tuple[type, ...], data:dict[str,Any]):
        super(SingletonMeta, cls).__init__(name, bases, data)
        if not hasattr(cls, "_instance"):
            cls._instance = None

    def __call__(cls, *args: Any) -> type:
        if cls._instance is None:
            logging.debug("Constructing Singleton: {}", cls.__name__)
            cls._instance = super().__call__(*args)

        return cls._instance

class SingletonMetaAlt(type(Protocol)):
    """
    Create an instance field to hold the singleton,
    Subclasses are separate singletons
    """

    def __init__(cls, name:str, bases:tuple[type, ...], data:dict[str,Any]):
        super(SingletonMetaAlt, cls).__init__(name, bases, data)
        cls._instance = None

    def __call__(cls, *args:Any) -> type:
        if cls._instance is None:
            logging.debug("Constructing Singleton: {}", cls.__name__)
            cls._instance = super().__call__(*args)

        return cls._instance

class PoolMeta(type(Protocol)):
    """
    Create an instance field to hold the singleton,
    Subclasses are separate singletons
    """
    _pool_size = 10

    def __init__(cls, name:str, bases:tuple[type, ...], data:dict[str,Any]):
        super(PoolMeta, cls).__init__(name, bases, data)
        cls._pool      : list[Any] = []
        cls._pool_last = 0

    def __call__(cls, *args:Any) -> type:
        if cls._pool_last > cls._pool_size - 1:
            cls._pool_last = 0

        if len(cls._pool) < cls._pool_size:
            logging.debug("Constructing Singleton Pool: {}", cls.__name__)
            obj = super().__call__(*args)
            cls._pool.append(obj)
        else:
            obj = cls._pool[cls._pool_last]
            obj.__init__(*args)

        cls._pool_last += 1
        return obj
